#!/usr/local/bin/python

import os
import sys
import time
import shutil
import datetime
import logging
import re
import tarfile
import stat
from texttestlib.default.batch import testoverview
from texttestlib import plugins
from .summarypages import GenerateSummaryPage, GenerateGraphs  # only so they become package level entities
from collections import OrderedDict
from .batchutils import getBatchRunName, BatchVersionFilter, parseFileName, convertToUrl
import subprocess
from glob import glob


class BatchCategory(plugins.Filter):
    def __init__(self, state):
        self.name = state.category
        self.briefDescription, self.longDescription = state.categoryDescriptions.get(self.name, (self.name, self.name))
        self.setsFailureCode = state.getExitCode()
        self.tests = {}
        self.testSuites = []

    def addTest(self, test):
        self.tests[test.getRelPath()] = test

    def getTestLine(self, test):
        postText = test.state.getTypeBreakdown()[1]
        if len(postText) > 0:
            postText = " : " + postText
            return test.getIndent() + "- " + test.paddedRepr() + postText + "\n"
        return test.getIndent() + "- " + repr(test) + postText + "\n"

    def size(self):
        return len(self.tests)

    def acceptsTestCase(self, test):
        return test.getRelPath() in self.tests

    def acceptsTestSuiteContents(self, suite):
        return not suite.isEmpty()

    def describeBrief(self, app):
        if self.size() > 0:
            filters = [self]
            suite = app.createExtraTestSuite(filters)
            self.testSuites.append(suite)
            return "The following tests " + self.longDescription + " : \n" + \
                   self.getTestLines(suite) + "\n"

    def getTestLines(self, test):
        if test.classId() == "test-case":
            realTest = self.tests[test.getRelPath()]
            return self.getTestLine(realTest)
        else:
            lines = test.getIndent() + "In " + repr(test) + ":\n"
            for subtest in test.testcases:
                lines += self.getTestLines(subtest)
            return lines

    def getAllTests(self):
        allTests = []
        for suite in self.testSuites:
            for test in suite.testCaseList():
                allTests.append(self.tests[test.getRelPath()])
        return allTests

    def describeFull(self):
        fullDescriptionString = self.getFullDescription()
        if fullDescriptionString:
            return "\nDetailed information for the tests that " + self.longDescription + " follows...\n" + fullDescriptionString
        else:
            return ""

    def getFreeTextData(self):
        data = OrderedDict()
        for test in self.getAllTests():
            freeText = test.state.freeText
            if freeText:
                if freeText not in data:
                    data[freeText] = []
                data[freeText].append(test)
        return list(data.items())

    def testOutput(self, test):
        return repr(test) + " (under " + test.getRelPath() + ")"

    def getFullDescription(self):
        fullText = ""
        for freeText, tests in self.getFreeTextData():
            fullText += "--------------------------------------------------------" + "\n"
            if len(tests) == 1:
                test = tests[0]
                fullText += "TEST " + repr(test.state) + " " + self.testOutput(test) + "\n"
            else:
                fullText += str(len(tests)) + " TESTS " + repr(tests[0].state) + "\n"
            fullText += freeText
            if not freeText.endswith("\n"):
                fullText += "\n"
            if len(tests) > 1:
                fullText += "\n"
                for test in tests:
                    fullText += "-- " + self.testOutput(test) + "\n"
        return fullText


class BatchApplicationData:
    def __init__(self, suite):
        self.suite = suite
        self.categories = {}
        self.errorCategories = []
        self.failureCategories = []
        self.successCategories = []

    def storeCategory(self, test):
        category = test.state.category
        if category not in self.categories:
            batchCategory = BatchCategory(test.state)
            if not test.state.hasResults():
                self.errorCategories.append(batchCategory)
            elif test.state.hasSucceeded():
                self.successCategories.append(batchCategory)
            else:
                self.failureCategories.append(batchCategory)
            self.categories[category] = batchCategory
        self.categories[category].addTest(test)

    def failureCount(self):
        return self.totalTests(self.failCategories())

    def successCount(self):
        return self.totalTests(self.successCategories)

    def failCategories(self):
        return self.errorCategories + self.failureCategories

    def triggersExitCode(self):
        return any((c.setsFailureCode and c.size() for c in self.failCategories()))

    def allCategories(self):
        return self.failCategories() + self.successCategories

    def testCount(self):
        return self.totalTests(self.allCategories())

    def totalTests(self, categoryList):
        count = 0
        for category in categoryList:
            count += category.size()
        return count

    def getFailuresBrief(self):
        contents = ""
        for category in self.failCategories():
            contents += category.describeBrief(self.suite.app)
        return contents

    def getSuccessBrief(self):
        contents = ""
        for category in self.successCategories:
            contents += category.describeBrief(self.suite.app)
        return contents

    def getDetails(self):
        contents = ""
        for category in self.allCategories():
            contents += category.describeFull()
        return contents


class EmailResponder(plugins.Responder):
    def __init__(self, optionMap, *args):
        plugins.Responder.__init__(self)
        self.runId = getBatchRunName(optionMap)
        self.batchAppData = OrderedDict()
        self.allApps = OrderedDict()

    def notifyComplete(self, test):
        if test.app.emailEnabled():
            if test.app not in self.batchAppData:
                self.addApplication(test)
            self.batchAppData[test.app].storeCategory(test)

    def getRootSuite(self, test):
        if test.parent:
            return self.getRootSuite(test.parent)
        else:
            return test

    def addApplication(self, test):
        rootSuite = self.getRootSuite(test)
        app = test.app
        self.batchAppData[app] = BatchApplicationData(rootSuite)
        self.allApps.setdefault(app.name, []).append(app)

    def notifyAllComplete(self):
        mailSender = MailSender(self.runId)
        for appList in list(self.allApps.values()):
            batchDataList = list(map(self.batchAppData.get, appList))
            mailSender.send(batchDataList)


sectionHeaders = ["Summary of all Unsuccessful tests",
                  "Details of all Unsuccessful tests", "Summary of all Successful tests"]


class MailSender:
    def __init__(self, runId=""):
        self.runId = runId
        self.diag = logging.getLogger("Mail Sender")

    def send(self, batchDataList):
        app = batchDataList[0].suite.app
        mailContents = self.makeContents(batchDataList)
        self.sendOrStoreMail(app, mailContents, self.useCollection(app), not self.hasExitCodeErrors(batchDataList))

    def makeContents(self, batchDataList, headerSection=True):
        app = batchDataList[0].suite.app
        mailTitle = self.getMailTitle(app, batchDataList)
        mailContents = self.createMailHeaderSection(
            mailTitle, app, batchDataList) if headerSection else mailTitle + "\n"
        if len(batchDataList) > 1:
            for batchData in batchDataList:
                mailContents += self.getMailTitle(app, [batchData]) + "\n"
            mailContents += "\n"
        if not self.isAllSuccess(batchDataList):
            mailContents += self.performForAll(app, batchDataList,
                                               BatchApplicationData.getFailuresBrief, sectionHeaders[0])
            mailContents += self.performForAll(app, batchDataList, BatchApplicationData.getDetails, sectionHeaders[1])
        if not self.isAllFailure(batchDataList):
            mailContents += self.performForAll(app, batchDataList,
                                               BatchApplicationData.getSuccessBrief, sectionHeaders[2])
        return mailContents

    def performForAll(self, app, batchDataList, method, headline):
        contents = headline + " follows...\n" + \
            "---------------------------------------------------------------------------------" + "\n"
        for resp in batchDataList:
            if len(batchDataList) > 1:
                if headline.find("Details") != -1 and not resp is batchDataList[0]:
                    contents += "---------------------------------------------------------------------------------" + "\n"
                contents += self.getMailTitle(app, [resp]) + "\n\n"
            contents += method(resp) + "\n"
        return contents

    def storeMail(self, app, mailContents):
        localFileName = "batchreport." + app.name + app.versionSuffix()
        collFile = os.path.join(app.writeDirectory, localFileName)
        self.diag.info("Adding mail report contents to " + collFile)

        with plugins.openForWrite(collFile) as collateFile:
            collateFile.write(mailContents)
            collateFile.close()

    def sendOrStoreMail(self, app, mailContents, useCollection=False, isAllSuccess=False):
        plugins.log.info("Creating batch report for application " + app.fullName() + " ...")
        if useCollection:
            self.storeMail(app, mailContents)
            plugins.log.info("File written.")
        else:
            if not isAllSuccess or app.getBatchConfigValue("batch_mail_on_failure_only") != "true":
                errorMessage = self.sendMail(app, mailContents)
                if errorMessage:
                    plugins.log.info("FAILED. Details follow:\n" + errorMessage.strip())
                else:
                    plugins.log.info("done.")
            else:
                plugins.log.info("not sent: all tests succeeded or had known bugs.")

    def exceptionOutput(self):
        exctype, value = sys.exc_info()[:2]
        from traceback import format_exception_only
        return "".join(format_exception_only(exctype, value))

    def sendMail(self, app, mailContents):
        smtpServer = app.getConfigValue("smtp_server")
        smtpUsername = app.getConfigValue("smtp_server_username")
        smtpPassword = app.getConfigValue("smtp_server_password")
        fromAddress = app.getBatchConfigValue("batch_sender")
        recipients = app.getBatchConfigValue("batch_recipients")

        if not recipients:
            return self.getNoRecipientsMessage(app)

        toAddresses = plugins.commasplit(recipients)

        import smtplib
        smtp = smtplib.SMTP()
        try:
            smtp.connect(smtpServer)
        except Exception:  # Can't use SMTPException, because this raises socket.error usually
            return "Could not connect to SMTP server at " + smtpServer + "\n" + self.exceptionOutput()
        if smtpUsername:
            try:
                smtp.login(smtpUsername, smtpPassword)
            except smtplib.SMTPException:
                return "Failed to login as '" + smtpUsername + "' to SMTP server at " + smtpServer + \
                    "\n" + self.exceptionOutput()
        try:
            smtp.sendmail(fromAddress, toAddresses, mailContents)
        except smtplib.SMTPException:
            return "Mail could not be sent\n" + self.exceptionOutput()
        smtp.quit()

    def getNoRecipientsMessage(self, app):
        message = "Attempted to send mail, but batch_recipients is empty."
        if app.getBatchConfigValue("batch_use_collection") == "true":
            message += "\nHaving batch_use_collection set to true, but no batch_recipients indicates an erroneous configuration."
        return message

    def createMailHeaderSection(self, title, app, batchDataList):
        if self.useCollection(app):
            return self.getMachineTitle(batchDataList) + "\n" + self.runId + "\n" + \
                title + "\n\n"  # blank line separating headers from body
        else:
            return self.createMailHeaderForSend(self.runId, title, app)

    def createMailHeaderForSend(self, runId, title, app):
        fromAddress = app.getBatchConfigValue("batch_sender")
        toAddress = app.getBatchConfigValue("batch_recipients")
        return "From: " + fromAddress + "\nTo: " + toAddress + "\n" + \
               "Subject: " + runId + " " + title + "\n\n"

    def useCollection(self, app):
        return app.getBatchConfigValue("batch_use_collection") == "true"

    def getMailHeader(self, app, batchDataList):
        versions = self.findCommonVersions(app, batchDataList)
        return app.fullName() + self.getVersionString(versions) + " : "

    def getCategoryNames(self, batchDataList):
        names = []
        for resp in batchDataList:
            for cat in resp.errorCategories:
                if not cat.name in names:
                    names.append(cat.name)
        for resp in batchDataList:
            for cat in resp.failureCategories:
                if not cat.name in names:
                    names.append(cat.name)
        for resp in batchDataList:
            for cat in resp.successCategories:
                if not cat.name in names:
                    names.append(cat.name)
        return names

    def hasExitCodeErrors(self, batchDataList):
        return any((bd.triggersExitCode() for bd in batchDataList))

    def isAllSuccess(self, batchDataList):
        return self.getTotalString(batchDataList, BatchApplicationData.failureCount) == "0"

    def isAllFailure(self, batchDataList):
        return self.getTotalString(batchDataList, BatchApplicationData.successCount) == "0"

    def getMailTitle(self, app, batchDataList):
        title = self.getMailHeader(app, batchDataList)
        title += self.getTotalString(batchDataList, BatchApplicationData.testCount) + " tests"
        if self.isAllSuccess(batchDataList):
            return title + ", all successful"
        title += " :"
        for categoryName in self.getCategoryNames(batchDataList):
            totalInCategory = self.getCategoryCount(categoryName, batchDataList)
            briefDesc = self.getBriefDescription(categoryName, batchDataList)
            title += self.briefText(totalInCategory, briefDesc)
        # Lose trailing comma
        return title[:-1]

    def getMachineTitle(self, batchDataList):
        values = []
        for categoryName in self.getCategoryNames(batchDataList):
            countStr = str(self.getCategoryCount(categoryName, batchDataList))
            briefDesc = self.getBriefDescription(categoryName, batchDataList)
            values.append(briefDesc + "=" + countStr)
        return ",".join(values)

    def getTotalString(self, batchDataList, method):
        total = 0
        for resp in batchDataList:
            total += method(resp)
        return str(total)

    def getCategoryCount(self, categoryName, batchDataList):
        total = 0
        for resp in batchDataList:
            if categoryName in resp.categories:
                total += resp.categories[categoryName].size()
        return total

    def getBriefDescription(self, categoryName, batchDataList):
        for resp in batchDataList:
            if categoryName in resp.categories:
                return resp.categories[categoryName].briefDescription

    def getVersionString(self, versions):
        if len(versions) > 0:
            return " " + ".".join(versions)
        else:
            return ""

    def briefText(self, count, description):
        if count == 0 or description == "succeeded":
            return ""
        else:
            return " " + str(count) + " " + description + ","

    def findCommonVersions(self, app, batchDataList):
        if len(batchDataList) == 0:
            return app.versions
        commonVersions = []
        otherBatchData = batchDataList[1:]
        for trialVersion in batchDataList[0].suite.app.versions:
            if self.allContain(otherBatchData, trialVersion):
                commonVersions.append(trialVersion)
        return commonVersions

    def allContain(self, otherBatchData, trialVersion):
        for batchData in otherBatchData:
            if not trialVersion in batchData.suite.app.versions:
                return False
        return True


def findExtraVersionParent(app, allApps):
    for parentApp in allApps:
        if app in parentApp.extras:
            return parentApp
    return app


def getVersionName(app, allApps):
    parent = findExtraVersionParent(app, allApps)
    parentVersion = parent.getFullVersion()
    fullVersion = app.getFullVersion()
    if parentVersion:
        return fullVersion
    elif fullVersion:
        return "default." + fullVersion
    else:
        return "default"


def getBatchRepository(suite):
    repo = suite.app.getBatchConfigValue("batch_result_repository", envMapping=suite.environment)
    return os.path.expanduser(repo)


def dateInSeconds(val):
    return time.mktime(time.strptime(val, "%d%b%Y"))


def matchesApp(dir, app):
    suffix = app.versionSuffix()
    return dir.startswith(app.name + suffix) or dir.startswith(app.getBatchSession() + suffix)


def getPreviousWriteDirsUnder(rootDir, app):
    dirs = []
    for dir in sorted(os.listdir(rootDir)):
        fullDir = os.path.join(rootDir, dir)
        if os.path.isdir(fullDir) and matchesApp(dir, app):
            dirs.append(fullDir)
    return dirs


def getPreviousWriteDirs(app):
    rootDir = app.getPreviousWriteDirInfo()
    return getPreviousWriteDirsUnder(rootDir, app) if os.path.isdir(rootDir) else []


def writeSuccessLine(f, runPostfix, state):
    f.write(runPostfix)
    if state.briefText:
        f.write(" " + state.briefText)
    f.write(" " + ", ".join(state.executionHosts) + "\n")


# Allow saving results to a historical repository
class SaveState(plugins.Responder):
    def __init__(self, optionMap, allApps):
        plugins.Responder.__init__(self)
        self.runPostfix = self.getRunPostfix(optionMap)
        self.failureFileName = "teststate_" + self.runPostfix
        self.successFileName = "succeeded_runs"
        self.repositories = {}
        self.allApps = allApps
        self.diag = logging.getLogger("Save Repository")
        self.checkRunNameValid()

    def isBatchDate(self, dateStr):
        return re.match("^[0-9]{2}[A-Za-z]{3}[0-9]{4}$", dateStr)

    def getRunPostfix(self, optionMap):
        # include the date and the name, if any. Date is used for archiving, name for display
        runName = getBatchRunName(optionMap)
        parts = []
        if not self.isBatchDate(runName):
            parts.append(plugins.startTimeString("%d%b%Y"))
        parts.append(runName)
        return "_".join(parts)

    def notifyComplete(self, test):
        if test.state.isComplete():  # might look weird but this notification also comes in scripts, e.g collecting
            test.saveState()
            if test.app in self.repositories:
                self.diag.info("Saving " + repr(test) + " to repository")
                self.saveToRepository(test)
            else:
                self.diag.info("No repositories for " + repr(test.app) + " in " + repr(self.repositories))

    def saveToRepository(self, test):
        testRepository = self.repositories[test.app]
        targetDir = os.path.join(testRepository, test.app.name, getVersionName(test.app, self.allApps),
                                 test.getRelPath())
        try:
            plugins.ensureDirectoryExists(targetDir)
        except EnvironmentError:
            plugins.printWarning("Could not create directory at " + targetDir)
        # Need to store teststate files for succeeded tests if we have resource pages
        if test.state.hasSucceeded() and len(test.app.getBatchConfigValue("historical_report_resources")) == 0:
            targetFile = os.path.join(targetDir, self.successFileName)
            with open(targetFile, "a") as f:
                writeSuccessLine(f, self.runPostfix, test.state)
        else:
            targetFile = os.path.join(targetDir, self.failureFileName)
            if os.path.isfile(targetFile):
                plugins.printWarning("File already exists at " + targetFile + " - not overwriting!")
            else:
                try:
                    shutil.copyfile(test.getStateFile(), targetFile)
                except EnvironmentError:
                    plugins.printWarning("Could not write file at " + targetFile)

    def addSuite(self, suite):
        testStateRepository = getBatchRepository(suite)
        self.diag.info("Test state repository is " + repr(testStateRepository))
        if testStateRepository:
            self.repositories[suite.app] = os.path.abspath(testStateRepository)
            self.registerRunName(testStateRepository, suite.app)

    def makeInitialRunFile(self, app, runFile):
        plugins.ensureDirExistsForFile(runFile)
        with open(runFile, "w") as f:
            for envVar in ["JENKINS_URL", "JOB_NAME", "BUILD_NUMBER"]:
                if envVar in os.environ:
                    f.write(envVar + "=" + os.getenv(envVar) + "\n")
                else:
                    break

            f.write(repr(app) + "\n")

    def runAlreadyExists(self, runFile, app):
        text = repr(app)
        with open(runFile) as f:
            for line in f:
                if line.strip() == text:
                    return True
        return False

    def getRunFileName(self, repository):
        return os.path.join(repository, "run_names", self.runPostfix)

    def registerRunName(self, repository, app):
        runFile = self.getRunFileName(repository)
        if os.path.isfile(runFile):
            with open(runFile, "a") as f:
                f.write(repr(app) + "\n")
        else:
            try:
                self.makeInitialRunFile(app, runFile)
            except EnvironmentError:
                plugins.printWarning("Could not write run file at " + runFile)

    def checkRunNameValid(self):
        for app in self.allApps:
            repository = os.path.expanduser(app.getBatchConfigValue("batch_result_repository"))
            runFile = self.getRunFileName(repository)
            if os.path.isfile(runFile) and self.runAlreadyExists(runFile, app):
                raise plugins.TextTestError("ERROR: Cannot run batch tests with run name '" + self.runPostfix + "', name has already been used for a different run\n" +
                                            "See file at " + runFile + ", which contains the entry '" + repr(app) + "'")


class MigrateBatchRepository(plugins.Action):
    def __init__(self):
        self.successFileName = "succeeded_runs"

    def migrateFile(self, path):
        state = testoverview.GenerateWebPages.readState(path)
        if state.hasSucceeded():
            dirname, fn = os.path.split(path)
            runTag = fn.replace("teststate_", "")
            successFile = os.path.join(dirname, self.successFileName)
            with open(successFile, "a") as f:
                writeSuccessLine(f, runTag, state)
            os.remove(path)

    def migrate(self, repository):
        plugins.log.info("Migrating repository at " + repository)
        for root, _, files in os.walk(repository):
            for f in files:
                if f.startswith("teststate_"):
                    path = os.path.join(root, f)
                    self.migrateFile(path)

    def setUpSuite(self, suite):
        if suite.parent is None:
            resources = suite.app.getBatchConfigValue("historical_report_resources")
            if len(resources) > 0:
                raise plugins.TextTestError("Cannot migrate repository: historical_report_sources set to '" +
                                            ", ".join(resources) + "' - these require the full format")
            repository = getBatchRepository(suite)
            if not os.path.isdir(repository):
                raise plugins.TextTestError("Batch result repository " + repository + " does not exist")
            self.migrate(repository)


class ArchiveScript(plugins.ScriptWithArgs):
    def __init__(self, argDict):
        self.descriptors = []
        self.beforeDate = self.parseDate(argDict, "before")
        self.afterDate = self.parseDate(argDict, "after")
        self.descriptors.sort()
        self.repository = None
        if len(argDict) == 0:
            raise plugins.TextTestError("Cannot archive the entire repository - give cutoff dates or run names!")

    def parseDate(self, dict, key):
        return self.extractArg(dict, key, isDate=True)

    def extractArg(self, dict, key, isDate=False):
        if not key in dict:
            return
        val = dict[key]
        if isDate:
            self.descriptors.append("dated " + key + " " + val)
            return self.dateInSeconds(val)
        else:
            self.descriptors.append("with " + key + " " + val)
            return val

    def getDateFormat(self, val):
        return "%b%Y" if len(val) == 7 else "%d%b%Y"

    def dateInSeconds(self, val):
        dateFormat = self.getDateFormat(val)
        return time.mktime(time.strptime(val, dateFormat))

    def getShortDescriptorText(self):
        shortDescriptors = [d.split(" ", 1)[-1] for d in self.descriptors]
        return "_".join(shortDescriptors).replace(" ", "_")

    def makeTarArchive(self, suite, repository):
        historyDir = suite.app.name + "_history"
        fullHistoryDir = os.path.join(repository, historyDir)
        if os.path.isdir(fullHistoryDir):
            tarFileName = historyDir + "_" + self.getShortDescriptorText() + ".tar.gz"
            plugins.log.info("Archiving completed for " + self.repository + ", created tarfile at " + tarFileName)
            subprocess.call(["tar", "cfz", tarFileName, historyDir], cwd=repository)
            plugins.rmtree(fullHistoryDir)

    def archiveFiles(self, suite):
        self.archiveFilesUnder(self.repository, suite.app)

    def archiveFilesUnder(self, repository, app, *args):
        count = 0
        dirList = os.listdir(repository)
        dirList.sort()
        for file in dirList:
            fullPath = os.path.join(repository, file)
            if self.shouldArchive(file, *args):
                if self.archiveFile(fullPath, app, *args):
                    count += 1
            elif os.path.isdir(fullPath):
                self.archiveFilesUnder(fullPath, app, *args)
        if count > 0:
            plugins.log.info("Archived " + str(count) + " files " + ", ".join(self.descriptors) +
                             " under " + repository.replace(self.repository + os.sep, ""))

    def setUpSuite(self, suite):
        if suite.parent is None:
            repository = self.getRepository(suite)
            self.repository = os.path.join(repository, suite.app.name)
            if not os.path.isdir(self.repository):
                raise plugins.TextTestError("Batch result repository " + self.repository + " does not exist")
            self.archiveFiles(suite)
            if os.name == "posix":
                self.makeTarArchive(suite, repository)

    def archiveFile(self, fullPath, app, *args):
        targetPath = self.getTargetPath(fullPath, app.name)
        plugins.ensureDirExistsForFile(targetPath)
        try:
            os.rename(fullPath, targetPath)
            return True
        except EnvironmentError as e:
            plugins.log.info("Rename failed: " + fullPath + " " + targetPath)
            plugins.log.info("Error was " + str(e))
            return False

    def getTargetPath(self, fullPath, appName):
        parts = fullPath.split(os.sep)
        parts.reverse()
        appIndex = parts.index(appName)
        parts[appIndex] = appName + "_history"
        parts.reverse()
        return os.sep.join(parts)

    def shouldArchiveWithDate(self, date):
        if self.beforeDate and date >= self.beforeDate:
            return False
        if self.afterDate and date <= self.afterDate:
            return False
        return True


class ArchiveRepository(ArchiveScript):
    scriptDoc = "Archive parts of the batch result repository to a history directory"

    def __init__(self, args):
        argDict = self.parseArguments(args, ["before", "after", "weekday_pages_before", "name"])
        ArchiveScript.__init__(self, argDict)
        self.name = self.extractArg(argDict, "name")
        self.weekdayBeforeDate = self.parseDate(argDict, "weekday_pages_before")

    def getRepository(self, suite):
        return getBatchRepository(suite)

    def getWeekDays(self, suite):
        weekdayNameLists = list(suite.getConfigValue("historical_report_subpage_weekdays").values())
        weekdayNames = sum(weekdayNameLists, [])
        return list(map(plugins.weekdays.index, weekdayNames))

    def archiveFiles(self, suite):
        weekdays = self.getWeekDays(suite)
        runDir = os.path.join(os.path.dirname(self.repository), "run_names")
        if os.path.isdir(runDir):
            self.archiveFilesUnder(runDir, suite.app, weekdays, True)
        self.archiveVersionDirsUnder(self.repository, suite.app, weekdays, False)

    def archiveVersionDirsUnder(self, repository, app, *args):
        appVersions = set(app.versions)
        for directory in sorted(os.listdir(repository)):
            dirversions = set(directory.split("."))
            fullPath = os.path.join(repository, directory)
            if appVersions.issubset(dirversions) and os.path.isdir(fullPath):
                self.archiveFilesUnder(fullPath, app, *args)

    def archiveRunFile(self, fullPath, app):
        appParts = set(repr(app).split("."))
        envVars, otherAppLines = [], []
        archived = False
        with open(fullPath) as f:
            for line in f:
                if "=" in line:
                    envVars.append(line)
                else:
                    lineParts = set(line.strip().split("."))
                    if appParts.issubset(lineParts):
                        archived = True
                    else:
                        otherAppLines.append(line)
        if archived:
            if otherAppLines:
                with open(fullPath, "w") as f:
                    for line in envVars + otherAppLines:
                        f.write(line)
            else:
                os.remove(fullPath)
        return archived

    def archiveFile(self, fullPath, app, weekdays, isRunFile):
        if isRunFile:
            return self.archiveRunFile(fullPath, app)
        elif os.path.basename(fullPath).startswith("teststate"):
            return ArchiveScript.archiveFile(self, fullPath, app)
        linesToKeep = []
        archived = False
        with open(fullPath) as readFile:
            for line in readFile:
                tag = line.strip().split()[0]
                if self.shouldArchiveGivenTag(tag, weekdays):
                    archived = True
                else:
                    linesToKeep.append(line)

        if archived:
            with open(fullPath, "w") as writeFile:
                for line in linesToKeep:
                    writeFile.write(line)
        return archived

    def shouldArchiveGivenTag(self, tag, weekdays):
        parts = tag.split("_")
        dateStr = parts[0]
        if self.name:
            runName = dateStr if len(parts) == 1 else "_".join(parts[1:])
            if runName != self.name:
                return False
        return self.shouldArchiveGivenDateString(dateStr, weekdays)

    def shouldArchive(self, file, weekdays, isRunFile):
        if isRunFile:
            return self.shouldArchiveGivenTag(file, weekdays)

        if file.startswith("succeeded_"):
            return True
        if not file.startswith("teststate"):
            return False

        return self.shouldArchiveGivenTag(file.replace("teststate_", ""), weekdays)

    def shouldArchiveGivenDateString(self, dateStr, weekdays):
        timeStruct = time.strptime(dateStr, self.getDateFormat(dateStr))
        date = time.mktime(timeStruct)
        return self.shouldArchiveWithDate(date) and (not self.weekdayBeforeDate or date < self.weekdayBeforeDate or timeStruct.tm_wday not in weekdays)


class ArchiveHTML(ArchiveScript):
    scriptDoc = "Archive parts of the historical report location to a history directory"

    def __init__(self, args):
        argDict = self.parseArguments(args, ["before", "after"])
        ArchiveScript.__init__(self, argDict)

    def getRepository(self, suite):
        return os.path.expanduser(suite.app.getBatchConfigValue("historical_report_location"))

    def parseDateFromFile(self, file):
        for part in reversed(file[:-5].split("_")):
            if len(part) in [7, 9]:
                try:
                    return self.dateInSeconds(part)
                except ValueError:
                    pass

    def shouldArchive(self, file):
        if not file.endswith(".html"):
            return False
        date = self.parseDateFromFile(file)
        return self.shouldArchiveWithDate(date) if date else False


class ArchiveExtractor:
    def __init__(self, dateStr):
        self.dateStr = dateStr
        self.repositories = []

    def extract(self, suite):
        repository = getBatchRepository(suite)
        archives = self.getArchives(suite)
        for archive in archives:
            self.extractUnder(archive, repository)
        self.repositories.append(os.path.join(repository, suite.app.name + "_history"))

    def getArchives(self, suite):
        repository = getBatchRepository(suite)
        dirList = os.listdir(repository)
        return [os.path.join(repository, f) for f in dirList if f.endswith(".tar.gz") and self.shouldOpen(f)]

    def shouldOpen(self, tarFileName):
        beforePos = tarFileName.rfind("before_")
        if beforePos == -1:
            return False

        startPos = beforePos + len("before_")
        dateStr = tarFileName[startPos:startPos + 9]  # %d%b%Y dates are always of length 9
        return dateInSeconds(self.dateStr) <= dateInSeconds(dateStr)

    def extractUnder(self, archivedFile, targetPath):
        tar = tarfile.open(archivedFile)
        tar.extractall(targetPath, members=self.findFiles(tar))
        tar.close()

    def findFiles(self, members):
        for tarinfo in members:
            if self.shouldExtract(os.path.split(tarinfo.name)[1]):
                yield tarinfo

    def shouldExtract(self, fileName):
        if not fileName.startswith("teststate"):
            return False
        dateStr = fileName.split("_")[1]
        return dateInSeconds(self.dateStr) <= dateInSeconds(dateStr)

    def cleanAllExtracted(self):
        for repository in self.repositories:
            if os.path.isdir(repository):
                plugins.rmtree(repository)


class WebPageResponder(plugins.Responder):
    def __init__(self, optionMap, allApps):
        plugins.Responder.__init__(self)
        self.archiveExtractor = ArchiveExtractor(optionMap.get(
            "collarchive")) if optionMap.get("collarchive") is not None else None
        self.diag = logging.getLogger("GenerateWebPages")
        self.suitesToGenerate = []
        self.archiveUnused = "manualarchive" not in optionMap
        self.descriptionInfo = {}
        self.summaryGenerator = GenerateSummaryPage()

    def notifyAdd(self, test, *args, **kw):
        self.descriptionInfo.setdefault(test.app, {}).setdefault(
            test.getRelPath().replace(os.sep, " "), test.description)

    def addSuites(self, suites):
        # Don't blanket remove rejected apps automatically when collecting
        self.extraApps = []
        for suite in suites:
            if suite.app in self.extraApps:
                continue
            if self.summaryGenerator:
                self.summaryGenerator.setUpApplication(suite.app)
            try:
                batchFilter = BatchVersionFilter(suite.app.getBatchSession())
                batchFilter.verifyVersions(suite.app)
                self.suitesToGenerate.append(suite)
                self.extraApps += suite.app.extras
                if self.archiveExtractor is not None:
                    self.extractFromArchive()
            except (plugins.TextTestError, plugins.TextTestWarning) as e:
                plugins.log.info("Not generating web page for " + suite.app.description() + " : " + str(e))

    def notifyAllRead(self, *args):
        # Sort the suites to generate by the number of tests they contain
        # This should ensure we get small test suites' results first and anything that is slow to generate
        # won't slow down other stuff
        self.suitesToGenerate.sort(key=lambda s: s.size())

    def notifyAllComplete(self):
        appInfo = self.getAppRepositoryInfo()
        plugins.log.info("Generating web pages...")
        for pageTitle, pageInfo in list(appInfo.items()):
            plugins.log.info("Generating page for " + pageTitle)
            if len(pageInfo) == 1:
                self.generatePagePerApp(pageTitle, pageInfo)
            else:
                self.generateCommonPage(pageTitle, pageInfo)

            if self.summaryGenerator:
                appsUsed = [app for app, _, _ in pageInfo]
                self.summaryGenerator.generateForApps(appsUsed)
        if len(appInfo) == 0 and self.summaryGenerator:
            # Describe errors, if any
            self.summaryGenerator.finalise()
        plugins.log.info("Completed web page generation.")
        if self.archiveExtractor is not None:
            self.archiveExtractor.cleanAllExtracted()

    def extractFromArchive(self):
        for suite in self.suitesToGenerate:
            self.archiveExtractor.extract(suite)

    def generatePagePerApp(self, pageTitle, pageInfo):
        for app, repositories, extraApps in pageInfo:
            pageTopDir = os.path.expanduser(app.getBatchConfigValue("historical_report_location"))
            pageDir = os.path.join(pageTopDir, app.name)
            self.copyJavaScript(pageTopDir, pageDir)
            extraVersions = self.getExtraVersions(app, extraApps)
            self.diag.info("Found extra versions " + repr(extraVersions))
            relevantSubDirs = self.findRelevantSubdirectories(repositories, app, extraVersions)
            version = getVersionName(app, self.getAppsToGenerate())
            pageSubTitles = self.makePageSubTitles([app])
            self.makeAndGenerate(relevantSubDirs, self.getConfigValueMethod(app), pageDir, pageTitle, pageSubTitles,
                                 version, extraVersions, self.getDescriptionInfo([app]))

    def getConfigValueMethod(self, app):
        def getConfigValue(key, subKey=app.getBatchSession(), allSubKeys=False):
            if allSubKeys:
                return app.getConfigValue(key)
            else:
                return app.getCompositeConfigValue(key, subKey)
        return getConfigValue

    def makePageSubTitles(self, apps):
        cmdLine = self.makeCommandLine(apps)
        startText = "To start TextTest for these tests, run:"
        subtitles = [(startText, cmdLine)]
        if len(apps) == 1:
            reconnectCmdLine = self.makeReconnectCommandLine(apps[0], cmdLine)
            if reconnectCmdLine:
                reconnectText = "To reconnect the TextTest GUI to these results, run:"
                subtitles.insert(0, (reconnectText, reconnectCmdLine))
        return subtitles

    def makeCommandLine(self, apps):
        cmd = os.path.basename(plugins.getTextTestProgram())
        cmd += " -a " + ",".join((app.name for app in apps))
        version = apps[0].getFullVersion()
        if version:
            cmd += " -v " + version
        checkouts = set((app.getCheckoutForDisplay() for app in apps))
        if len(checkouts) == 1:
            checkout = checkouts.pop()
            if checkout:
                cmd += " -c " + checkout
        directories = set((app.getRootDirectory() for app in apps))
        cmd += " -d " + os.pathsep.join(directories)
        return cmd

    def makeReconnectCommandLine(self, app, cmdLine):
        previousWriteDirs = getPreviousWriteDirs(app)
        if not previousWriteDirs:
            return

        return cmdLine + " -reconnect " + previousWriteDirs[0] + " -g"

    def getAppRepositoryInfo(self):
        appInfo = OrderedDict()
        for suite in self.suitesToGenerate:
            app = suite.app
            repositories = self.getRepositories(suite)
            if len(repositories) == 0:
                continue
            pageTitle = app.getBatchConfigValue("historical_report_page_name")
            extraApps = []
            for extraApp in app.extras:
                extraPageTitle = extraApp.getBatchConfigValue("historical_report_page_name")
                if extraPageTitle != pageTitle and extraPageTitle != extraApp.getDefaultPageName():
                    appInfo.setdefault(extraPageTitle, []).append((extraApp, repositories, []))
                else:
                    extraApps.append(extraApp)
            appInfo.setdefault(pageTitle, []).append((app, repositories, extraApps))
        return appInfo

    def getRepositories(self, suite):
        repositories = []
        repositoryRoot = getBatchRepository(suite)
        if repositoryRoot:
            repos = [os.path.join(repositoryRoot, suite.app.name)]
            if self.archiveExtractor is not None:
                repos.append(os.path.join(repositoryRoot, suite.app.name + "_history"))
            repositories = [repository for repository in repos if self.checkRepository(repository, suite.app)]
        return repositories

    def checkRepository(self, repository, app):
        if not os.path.isdir(repository):
            plugins.printWarning("Batch result repository " + repository +
                                 " does not exist - not creating pages for " + repr(app))
            return False
        return True

    def getAppsToGenerate(self):
        return [suite.app for suite in self.suitesToGenerate]

    def getDescriptionInfo(self, apps):
        descriptionInfo = {}
        for app in apps:
            for appToUse in [app] + app.extras:
                descriptionInfo.update(self.descriptionInfo.get(appToUse, {}))

        return descriptionInfo

    def transformToCommon(self, pageInfo):
        # We've sorted by number of tests already, but on the same page we want applications in a more predictable order
        # Sort by name again.
        pageInfo.sort(key=lambda info: info[0].name)
        allApps = [app for app, _, _ in pageInfo]
        version = getVersionName(allApps[0], self.getAppsToGenerate())
        extraVersions, relevantSubDirs = [], OrderedDict()
        for app, repositories, extraApps in pageInfo:
            extraVersions += self.getExtraVersions(app, extraApps)
            relevantSubDirs.update(self.findRelevantSubdirectories(
                repositories, app, extraVersions, self.getVersionTitle))
        getConfigValue = plugins.ResponseAggregator([self.getConfigValueMethod(app) for app in allApps])
        pageSubTitles = self.makePageSubTitles(allApps)
        descriptionInfo = self.getDescriptionInfo(allApps)
        return relevantSubDirs, getConfigValue, version, extraVersions, pageSubTitles, descriptionInfo

    def getVersionTitle(self, app, version):
        title = app.fullName()
        if len(version) > 0 and version != "default":
            title += " version " + version
        return title

    def generateCommonPage(self, pageTitle, pageInfo):
        relevantSubDirs, getConfigValue, version, extraVersions, pageSubTitles, descriptionInfo = self.transformToCommon(
            pageInfo)
        pageDir = os.path.expanduser(getConfigValue("historical_report_location"))
        self.copyJavaScript(pageDir, pageDir)
        self.makeAndGenerate(relevantSubDirs, getConfigValue, pageDir, pageTitle,
                             pageSubTitles, version, extraVersions, descriptionInfo)

    def copyJavaScript(self, pageTopDir, pageDir):
        jsDir = os.path.join(pageTopDir, "javascript")
        srcDir = os.path.join(os.path.dirname(__file__), "testoverview_javascript")
        if not os.path.isdir(srcDir): # for example when zipped, use the installation structure
            moduleParts = self.__class__.__module__.split(".")
            pathParts = moduleParts[1:] +  [ "testoverview_javascript" ] # drop texttestlib at the start, installation is under that
            srcDir = plugins.installationPath(*pathParts)
        plugins.ensureDirectoryExists(jsDir)
        plugins.ensureDirectoryExists(pageDir)
        for fn in sorted(os.listdir(srcDir)):
            targetDir = pageDir if fn.endswith(".php") else jsDir  # Server-side stuff is per application
            targetPath = os.path.join(targetDir, fn)
            try:
                shutil.copyfile(os.path.join(srcDir, fn), targetPath)
            except EnvironmentError:
                plugins.printWarning("Could not copy Javascript-related file to " + targetPath)

    def makeAndGenerate(self, subDirs, getConfigValue, pageDir, *args):
        resourceNames = getConfigValue("historical_report_resources")
        plugins.ensureDirectoryExists(pageDir)
        try:
            self.generateWebPages(subDirs, getConfigValue, pageDir, resourceNames, *args)
        except Exception:  # pragma: no cover - robustness only, shouldn't be reachable
            sys.stderr.write("Caught exception while generating web pages :\n")
            plugins.printException()

    def getWebPageGenerator(self, getConfigValue, *args):
        return testoverview.GenerateWebPages(getConfigValue, *args)

    def generateWebPages(self, subDirs, getConfigValue, *args):
        generator = self.getWebPageGenerator(getConfigValue, *args)
        subPageNames = getConfigValue("historical_report_subpages")
        generator.generate(subDirs, subPageNames, self.archiveUnused)

    def findMatchingExtraVersion(self, dirVersions, extraVersions):
        # Check all tails that this is not an extraVersion
        for pos in range(len(dirVersions)):
            versionToCheck = ".".join(dirVersions[pos:])
            if versionToCheck in extraVersions:
                return versionToCheck
        return ""

    def findRelevantSubdirectories(self, repositories, app, extraVersions, versionTitleMethod=None):
        subdirs = OrderedDict()
        for repository in repositories:
            dirlist = os.listdir(repository)
            dirlist.sort()
            appVersions = set(app.versions)
            for dir in dirlist:
                dirVersions = dir.split(".")
                if set(dirVersions).issuperset(appVersions):
                    currExtraVersion = self.findMatchingExtraVersion(dirVersions, extraVersions)
                    if currExtraVersion:
                        version = dir.replace("." + currExtraVersion, "")
                    else:
                        version = dir
                    if versionTitleMethod:
                        versionTitle = versionTitleMethod(app, version)
                    else:
                        versionTitle = version
                    fullPath = os.path.join(repository, dir)
                    self.diag.info("Found subdirectory " + dir + " with version " + versionTitle
                                   + " and extra version '" + currExtraVersion + "'")
                    subdirs.setdefault(versionTitle, []).append((currExtraVersion, fullPath))
        return subdirs

    def getExtraVersions(self, app, extraApps):
        extraVersions = []
        length = len(app.versions)
        for extraApp in extraApps:
            version = ".".join(extraApp.versions[length:])
            if not version in app.versions:
                extraVersions.append(version)
        return extraVersions


class CollectFilesResponder(plugins.Responder):
    def __init__(self, optionMap, allApps):
        plugins.Responder.__init__(self)
        self.mailSender = MailSender()
        self.runId = ""  # depends on what we pick up from collected files
        self.diag = logging.getLogger("batch collect")
        self.allApps = allApps

    def notifyAllComplete(self):
        plugins.log.info("Collecting batch files locally...")
        for app in self.allApps:
            self.collectFilesForApp(app)

    def getMailBody(self, app, fileBodies, missingVersions):
        reportLocation = app.getBatchConfigValue("historical_report_location")
        if reportLocation:
            htmlIndex, htmlBody = self.getHtmlReportLocation(app, reportLocation)
            if htmlIndex is not None:
                return "Please see detailed results at " + htmlBody + "\n\n" + \
                    "The main index page can be found at " + htmlIndex + "\n"

        return self.getBody(fileBodies, missingVersions)

    def collectFilesForApp(self, app):
        fileBodies = []
        totalValues = OrderedDict()
        compulsoryVersions = set(app.getBatchConfigValue("batch_collect_compulsory_version"))
        versionsFound = set()
        rootDir = app.getPreviousWriteDirInfo()
        if not os.path.isdir(rootDir):
            sys.stderr.write("No temporary directory found at " + rootDir + " - not collecting batch reports.\n")
            return

        for previousWriteDir in getPreviousWriteDirsUnder(rootDir, app):
            currBodies, currVersions = self.parseDirectory(previousWriteDir, app, totalValues)
            fileBodies += currBodies
            versionsFound.update(currVersions)

        if len(fileBodies) == 0:
            return

        missingVersions = compulsoryVersions.difference(versionsFound)

        mailTitle = self.getTitle(app, totalValues)
        mailContents = self.mailSender.createMailHeaderForSend(self.runId, mailTitle, app)
        mailContents += self.getMailBody(app, fileBodies, missingVersions)
        allCats = set(totalValues.keys())
        noMailCats = set(["succeeded", "known bugs"])
        allSuccess = allCats.issubset(noMailCats)
        self.mailSender.sendOrStoreMail(app, mailContents, isAllSuccess=allSuccess)

    def getHtmlReportLocation(self, app, reportLocation):
        latestFile = self.getMostRecentFile(app, reportLocation)
        if latestFile is not None:
            fileMapping = app.getConfigValue("file_to_url")
            indexFile = os.path.join(reportLocation, "index.html")
            return convertToUrl(indexFile, fileMapping), convertToUrl(latestFile, fileMapping)
        else:
            return None, None

    def getMostRecentFile(self, app, reportLocation):
        versionName = getVersionName(app, self.allApps)
        appDir = os.path.join(reportLocation, app.name)
        bestPath, bestDate, bestTag = None, None, None
        for path in glob(os.path.join(appDir, "test_" + versionName + "_*.html")):
            fileName = os.path.basename(path)
            _, date, tag = parseFileName(fileName, self.diag)
            if date is not None:
                paddedTag = plugins.padNumbersWithZeroes(tag)
                if bestPath is None or date > bestDate or (date == bestDate and paddedTag > bestTag):
                    bestPath, bestDate, bestTag = path, date, paddedTag
        return bestPath

    def parseDirectory(self, fullDir, app, totalValues):
        basicPrefix = "batchreport." + app.name
        prefix = basicPrefix + app.versionSuffix()
        # Don't collect to more collections!
        self.diag.info("Setting up application " + app.name + " looking for " + prefix)
        filelist = os.listdir(fullDir)
        filelist.sort()
        fileBodies = []
        versionsFound = set()
        for filename in filelist:
            if filename == prefix or filename.startswith(prefix + "."):
                fullname = os.path.join(fullDir, filename)
                fileBody = self.parseFile(fullname, app, totalValues)
                if fileBody:
                    fileBodies.append(fileBody)
                    versionsFound.update(set(filename.replace(basicPrefix, "").split(".")))

        return fileBodies, versionsFound

    @staticmethod
    def runIsRelevant(runId, maxDays):
        if maxDays >= 100000:  # Default value
            return True
        try:
            runDate = datetime.date.fromtimestamp(time.mktime(time.strptime(runId, "%d%b%Y")))
        except ValueError:
            return True  # Isn't necessarily a date, in which case we have no grounds for rejecting it
        todaysDate = datetime.date.today()
        timeElapsed = todaysDate - runDate
        return timeElapsed.days <= maxDays

    def parseFile(self, fullname, app, totalValues):
        localName = os.path.basename(fullname)
        plugins.log.info("Found file called " + localName)
        file = open(fullname)
        valuesLine = file.readline()
        runId = file.readline().strip()
        maxDays = app.getBatchConfigValue("batch_collect_max_age_days")
        if self.runIsRelevant(runId, maxDays):
            self.runId = runId
            self.addValuesToTotal(localName, valuesLine, totalValues)
            fileBody = self.runId + " " + file.read()
            file.close()
            return fileBody
        else:
            plugins.log.info("Not including " + localName + " as run is more than " +
                             str(maxDays) + " days old (as determined by batch_collect_max_age_days).")

    def addValuesToTotal(self, localName, valuesLine, totalValues):
        catValues = plugins.commasplit(valuesLine.strip())
        try:
            for value in catValues:
                catName, count = value.split("=")
                if catName not in totalValues:
                    totalValues[catName] = 0
                totalValues[catName] += int(count)
        except ValueError:
            plugins.printWarning("Found truncated or old format batch report (" +
                                 localName + ") - could not parse result correctly.")

    def getTitle(self, app, totalValues):
        title = self.mailSender.getMailHeader(app, [])
        total = 0
        for value in list(totalValues.values()):
            total += value
        title += str(total) + " tests ran"
        if len(list(totalValues.keys())) == 1:
            return title + ", all " + list(totalValues.keys())[0]
        title += " :"
        for catName, count in list(totalValues.items()):
            title += self.mailSender.briefText(count, catName)
        # Lose trailing comma
        return title[:-1]

    def extractHeader(self, body):
        firstSep = body.find("\n") + 1
        header = body[0:firstSep]
        return header, body[firstSep:]

    def extractSection(self, sectionHeader, body):
        headerLoc = body.find(sectionHeader)
        if headerLoc == -1:
            return body.strip(), ""
        nextLine = body.find("\n", headerLoc) + 1
        if body[nextLine] == "-":
            nextLine = body.find("\n", nextLine) + 1
        section = body[0:headerLoc].strip()
        newBody = body[nextLine:].strip()
        return section, newBody

    def getBody(self, bodies, missingVersions):
        totalBody = ""
        for version in sorted(missingVersions):
            totalBody += "ERROR : No sufficiently recent run matching compulsory version '" + version + "' was found.\n"
        if len(bodies) == 1:
            return totalBody + bodies[0]

        parsedBodies = []
        for subBody in bodies:
            header, parsedSubBody = self.extractHeader(subBody)
            totalBody += header
            parsedBodies.append((header, parsedSubBody))
        totalBody += "\n"

        prevSectionHeader = ""
        for sectionHeader in sectionHeaders:
            parsedSections = []
            newParsedBodies = []
            for header, body in parsedBodies:
                section, newBody = self.extractSection(sectionHeader, body)
                if len(newBody) != 0:
                    newParsedBodies.append((header, newBody))
                if len(section) != 0:
                    parsedSections.append((header, section))

            totalBody += self.getSectionBody(prevSectionHeader, parsedSections)
            parsedBodies = newParsedBodies
            prevSectionHeader = sectionHeader
        totalBody += self.getSectionBody(prevSectionHeader, parsedBodies)
        return totalBody

    def getSectionBody(self, sectionHeader, parsedSections):
        if len(sectionHeader) == 0 or len(parsedSections) == 0:
            return ""
        sectionBody = sectionHeader + " follows...\n"
        detailSection = sectionHeader.find("Details") != -1
        if not detailSection or len(parsedSections) == 1:
            sectionBody += "=================================================================================\n"
        for header, section in parsedSections:
            if len(parsedSections) > 1:
                if detailSection:
                    sectionBody += "=================================================================================\n"
                sectionBody += header + "\n"
            sectionBody += section + "\n\n"
        return sectionBody



import sys
import datetime
import shutil
import time
import os
from collections import OrderedDict
from gi.repository import Gtk
from texttestlib import plugins
from .. import guiutils
from . import vcs_independent

#
# Todo/improvements:
#
# + Multiple dialogs confuses StoryText - close doesn't work correctly, for example ..
# + There is a lot of string stripping/replacing going on - perhaps this
#   could be unified and collected in a more centralized place?
# + Update on non-cvs controlled tests give no hint that tests are not in cvs.
# + Test:
#   - For update, we want a 'C ' file ...
#   - For add we need some new files and dirs ... and something that can cause
#     'add aborted' ...
#   - Remove needs some removed files ... and at least one file
#     which is re-born before pressing OK.
#   - Commit needs modified, added and removed files. We also want some error
#     which can cause the commit to fail, e.g. a simultaneous commit from someone else.
# + Commit
#   -r rev      Commit to this branch or trunk revision?
# + Fix update
#   - cvsrevertlast (cvs up -j <old> -j <new> ? (/usr/bin/cvsrevertlast))
#   - Non-modifying version should be available (-n) Maybe not necessary, cvs stat handles this nicely ...
#   - Overwrite local modifications should be available (-C)
#   -d
#   -P      Prune empty directories.
#   - Other date/revision options. Should we care?
# + What happens when we add a test which already has a CVS dir? (e.g.
#   when a test case/suite has been copied with plain 'cp -r' ...
#


class CVSInterface(vcs_independent.VersionControlInterface):
    def __init__(self, cvsDir):
        # Googled up.
        cvsWarningStates = ["Locally Modified", "Locally Removed", "Locally Added"]
        cvsErrorStates = ["File had conflicts on merge", "Needs Checkout", "Unresolved Conflicts", "Needs Patch",
                          "Needs Merge", "Entry Invalid", "Unknown", "PROHIBITED"]
        vcs_independent.VersionControlInterface.__init__(self, cvsDir, "CVS", cvsWarningStates, cvsErrorStates, "HEAD")
        self.defaultArgs["log"] = ["-N"]
        self.defaultArgs["diff"] = ["-N"]
        self.defaultArgs["rm"] = ["-f"]
        self.defaultArgs["update"] = ["-dP"]
        self.programArgs, self.errorMessage = self.setProgramArgs(cvsDir)

    def getProgramArgs(self):
        if self.errorMessage:
            raise plugins.TextTestError(self.errorMessage)
        else:
            return self.programArgs

    def setProgramArgs(self, cvsDir):
        cvsRoot = os.getenv("CVSROOT")
        if cvsRoot:
            return ["cvs", "-q"], ""
        else:
            rootFile = os.path.join(cvsDir, "Root")
            if os.path.isfile(rootFile):
                cvsRoot = self.getCvsRootFromFile(rootFile)
                return ["cvs", "-q", "-d", cvsRoot], ""
            else:
                return [], "Could not determine $CVSROOT: environment variable not set and no file present at:\n" + rootFile

    def getCvsRootFromFile(self, rootFile):
        info = open(rootFile).read()
        return info.strip().rstrip(os.sep)

    def getDateFromLog(self, output):
        for line in output.splitlines():
            if line.startswith("date:"):
                return datetime.datetime(*(self.parseCvsDateTime(line[6:25])[0:6]))

    def parseCvsDateTime(self, input):
        # Different CVS versions produce different formats...
        try:
            return time.strptime(input, "%Y/%m/%d %H:%M:%S")
        except ValueError:
            return time.strptime(input, "%Y-%m-%d %H:%M:%S")

    def getStateFromStatus(self, output):
        for line in output.splitlines():
            if line.startswith("File: "):
                spaceAfterNamePos = line.find("\t", 7)
                return line[spaceAfterNamePos:].replace("Status: ", "").strip(" \n\t")

    def isVersionControlled(self, path):
        if os.path.isdir(path):
            return os.path.isdir(os.path.join(path, "CVS"))
        else:
            return vcs_independent.VersionControlInterface.isVersionControlled(self, path)

    # Move in source control also. In CVS this implies a remove and then an add
    def _movePath(self, oldPath, newPath):
        self.checkInstalled()  # throws if it isn't, avoid moving paths around
        self.copyPath(oldPath, newPath)
        self.removePath(oldPath)
        self.callProgramOnFiles("add", newPath, recursive=True)

    def getMoveCommand(self):
        return "cvs rm' and 'cvs add"

    def copyPath(self, oldPath, newPath):
        createNewDir = os.path.isdir(oldPath) and not os.path.exists(newPath)
        vcs_independent.VersionControlInterface.copyPath(self, oldPath, newPath)
        if createNewDir:
            self.cleanControlDirs(newPath)

    def removePath(self, path):
        if os.path.isdir(path):
            retCode = self.callProgram("rm", [path])
        else:
            # removing a file affects the directory it lives in, whereas removing a directory shouldn't
            # affect the parent...
            retCode = self.callProgram("rm", [path], cwd=os.path.dirname(path))
        if retCode > 0:
            # Wasn't in version control, probably
            return plugins.removePath(path)
        else:
            self.cleanUnknownFiles(path)
            return True

    def cleanUnknownFiles(self, oldDir):
        for root, dirs, files in os.walk(oldDir):
            if "CVS" in dirs:
                dirs.remove("CVS")
            for file in files:
                os.remove(os.path.join(root, file))

    def cleanControlDirs(self, newDir):
        for root, dirs, files in os.walk(newDir):
            if "CVS" in dirs:
                dirs.remove("CVS")
                shutil.rmtree(os.path.join(root, "CVS"))

    # CVS add doesn't implicitly add directories, and it modifies its control dirs which are spread
    # throughout the tree (hence we need to change cwd so our traffic mechanism picks this up when testing)
    def callProgramOnFiles(self, cmdName, fileArg, recursive=False, extraArgs=[], **kwargs):
        if cmdName == "add":
            basicArgs = self.getCmdArgs(cmdName, extraArgs)
            for fileName in self.getFileNames(fileArg, recursive, includeDirs=True):
                self.callProgramWithHandler(fileName, basicArgs + [fileName], cwd=os.path.dirname(fileName), **kwargs)
        else:
            vcs_independent.VersionControlInterface.callProgramOnFiles(
                self, cmdName, fileArg, recursive, extraArgs, **kwargs)


class CVSLogLatest(vcs_independent.LogGUI):
    def __init__(self, *args):
        vcs_independent.LogGUI.__init__(self, *args)
        self.cmdName = "log"

    def getExtraArgs(self):
        return ["-rHEAD"]

    def _getTitle(self):
        return "Log Latest"

    def getResultDialogMessage(self):
        cmdArgs = vcs_independent.vcs.getCmdArgs(self.cmdName, self.getExtraArgs())
        message = "Showing latest log entries for the CVS controlled files.\nCVS command used: " + " ".join(cmdArgs)
        if not self.recursive:
            message += "\nSubdirectories were ignored."
        return message

    def storeResult(self, fileName, output, test):
        # Each file has something like:
        #
        # RCS file ...
        # Working file:
        # head ...
        # ...
        # description ...
        # ------------
        # revision ...
        # date ...
        # <comments>
        # ============
        #
        # We only want to show the Working file and the stuff from ----- to ===== ...
        linesToShow = ""
        enabled = False
        for line in output.splitlines():
            if line.startswith("Working file"):
                linesToShow += "\nFile: " + os.path.basename(line[14:]) + "\n"
                continue
            if line.startswith("--------------------"):
                enabled = True
            elif line.startswith("===================="):
                linesToShow += "====================\n"
                enabled = False
            if enabled:
                linesToShow += line + "\n"
        self.pages.setdefault(test.uniqueName, "")
        self.pages[test.uniqueName] += linesToShow

    def addContents(self):
        self.pages = OrderedDict()
        self.runAndParse()
        self.vbox = Gtk.VBox()
        self.addHeader()
        self.addNotebook()

    def addHeader(self):
        message = self.getResultDialogMessage()
        if message:
            hbox = Gtk.HBox()
            icon = Gtk.STOCK_DIALOG_INFO
            hbox.pack_start(self.getStockIcon(icon), False, False, 0)
            hbox.pack_start(Gtk.Label(message), False, False, 0)
            alignment = Gtk.Alignment.new(0.0, 1.0, 1.0, 1.0)
            alignment.set_padding(5, 5, 0, 5)
            alignment.add(hbox)
            self.vbox.pack_start(alignment, False, False, 0)

    def addNotebook(self):
        notebook = Gtk.Notebook()
        notebook.set_scrollable(True)
        notebook.popup_enable()
        for label, content in list(self.pages.items()):
            buffer = Gtk.TextBuffer()
            # Encode to UTF-8, necessary for Gtk.TextView
            buffer.set_text(content)
            textView = Gtk.TextView.new_with_buffer(buffer)
            textView.set_editable(False)
            window = Gtk.ScrolledWindow()
            window.set_policy(Gtk.PolicyType.AUTOMATIC, Gtk.PolicyType.AUTOMATIC)
            window.add(textView)
            notebook.append_page(window, Gtk.Label(label=label))
        notebook.show_all()
        if len(notebook.get_children()) > 0:  # Resize to a nice-looking dialog window ...
            parentSize = self.topWindow.get_size()
            self.dialog.resize(int(parentSize[0] / 1.5), int(parentSize[0] / 2))
        self.vbox.pack_start(notebook, True, True, 0)
        self.dialog.vbox.pack_start(self.vbox, True, True, 0)


vcs_independent.vcsClass = CVSInterface


class RenameTest(vcs_independent.RenameTest):
    def handleExistingDirectory(self, dir):
        if os.listdir(dir) == ["CVS"]:
            # There is only a CVS control dir, i.e. it's probably been removed in CVS.
            # Revert it in CVS and continue
            shutil.rmtree(dir)
            dirname, local = os.path.split(dir)
            vcs_independent.vcs.callProgram("update", [local], cwd=dirname)
        else:
            vcs_independent.RenameTest.handleExistingDirectory(self, dir)


class FilteredDiffGUI(vcs_independent.FilteredDiffGUI):
    def __init__(self, *args):
        vcs_independent.FilteredDiffGUI.__init__(self, *args)
        self.cmdName = "update"

    def getTmpFileArgs(self, fileName, revision):
        revArgs = vcs_independent.vcs.getSingleRevisionOptions(revision) if revision else []
        return ["-p"] + revArgs + [fileName]

    def commandHadError(self, retcode, stderr, stdout):
        # Diff returns an error code for differences, not just for errors
        return retcode or (len(stderr) > 0 and len(stdout) == 0)


class FilteredDiffGUIRecursive(FilteredDiffGUI):
    recursive = True

#
# Configuration for the Interactive Actions
#


class InteractiveActionConfig(vcs_independent.InteractiveActionConfig):
    def diffClasses(self):
        return [vcs_independent.DiffGUI, vcs_independent.DiffGUIRecursive, FilteredDiffGUI, FilteredDiffGUIRecursive]

    def getInteractiveActionClasses(self, dynamic):
        return vcs_independent.InteractiveActionConfig.getInteractiveActionClasses(self, dynamic) + [CVSLogLatest]

    def getRenameTestClass(self):
        return RenameTest




import os
import shutil
import operator
import logging
import time
import datetime
from texttestlib import plugins
from glob import glob
from itertools import groupby

# Trawl around for a suitable dir to reconnect to if we haven't been told one
# A tangle of side-effects: we find the run directory when asked for the extra versions,
# (so we can provide further ones accordingly), find the application directory when asked to check sanity
# (so we can bail if it's not there) and store in self.reconnDir, ready to provide to the ReconnectTest action


class ReconnectConfig:
    runDirCache = {}
    datedVersions = set()

    def __init__(self, optionMap):
        self.fullRecalculate = "reconnfull" in optionMap
        self.diag = logging.getLogger("Reconnection")
        self.reconnectTmpInfo = optionMap.get("reconnect")
        self.reconnDir = None
        self.errorMessage = ""

    def getReconnectAction(self):
        return ReconnectTest(self.reconnDir, self.fullRecalculate)

    def cacheRunDir(self, app, runDir, version=""):
        if version:
            keys = [app.fullName() + "." + version]
        else:
            keys = [app.fullName()] + app.versions
        for i in range(len(keys)):
            subKey = ".".join(keys[:i+1])
            if i == len(keys) - 1 or subKey not in self.runDirCache:
                self.runDirCache[subKey] = runDir
                self.diag.info("Caching " + subKey + " = " + runDir)

    def findRunDir(self, app):
        return self._findRunDir(repr(app))

    def _findRunDir(self, searchKey):
        self.diag.info("Searching for run directory for " + searchKey)
        entry = self.runDirCache.get(searchKey)
        if entry:
            return entry
        parts = searchKey.split(".")
        if len(parts) > 1:
            return self._findRunDir(".".join(parts[:-1]))

    def getExtraVersions(self, app, givenExtras):
        self.diag = logging.getLogger("Reconnection")
        self.diag.info("Finding reconnect 'extra versions' for " + repr(app) +
                       " given tmp info '" + repr(self.reconnectTmpInfo) + "'")
        if self.reconnectTmpInfo and os.path.isdir(self.reconnectTmpInfo):
            # See if this is an explicitly provided run directory
            dirName = os.path.normpath(self.reconnectTmpInfo)
            versionLists = self.getVersionListsTopDir(dirName)
            self.diag.info("Directory has version lists " + repr(versionLists))
            if versionLists is not None:
                return self.getVersionsFromDirs(app, [dirName], givenExtras)

        fetchDir = app.getPreviousWriteDirInfo(self.reconnectTmpInfo)
        if not os.path.isdir(fetchDir):
            if fetchDir == self.reconnectTmpInfo or not self.reconnectTmpInfo:
                self.errorMessage = "Could not find TextTest temporary directory at " + fetchDir
            else:
                self.errorMessage = "Could not find TextTest temporary directory for " + \
                                    self.reconnectTmpInfo + " at " + fetchDir
            return []

        self.diag.info("Looking for run directories under " + fetchDir)
        runDirs = self.getReconnectRunDirs(app, fetchDir)
        self.diag.info("Found run directories " + repr(runDirs))
        if len(runDirs) == 0:
            self.errorMessage = "Could not find any runs matching " + app.description() + " under " + fetchDir
            return []
        else:
            return self.getVersionsFromDirs(app, runDirs, givenExtras)

    def findAppDirUnder(self, app, runDir):
        # Don't pay attention to dated versions here...
        appVersions = frozenset(app.versions).difference(self.datedVersions)
        self.diag.info("Looking for directory with versions "
                       + repr(sorted(appVersions)))
        for f in os.listdir(runDir):
            versionSet = self.getVersionSetSubDir(f, app.name)
            if versionSet == appVersions:
                return os.path.join(runDir, f)

    def getReconnectRunDirs(self, app, fetchDir):
        correctNames = sorted(os.listdir(fetchDir))
        fullPaths = [os.path.join(fetchDir, d) for d in correctNames]
        return [d for d in fullPaths if self.isRunDirectoryFor(app, d)]

    def getFilter(self):
        return ReconnectFilter(self.reconnDir)

    @classmethod
    def all_perms(cls, items):
        # Lifted from a standard recipe
        if len(items) <= 1:
            yield items
        else:
            for perm in cls.all_perms(items[1:]):
                for i in range(len(perm)+1):
                    yield perm[:i] + items[0:1] + perm[i:]

    def versionSuffix(self, parts):
        fullVersion = ".".join(parts)
        if len(fullVersion) == 0:
            return ""
        return "." + fullVersion

    def isRunDirectoryFor(self, app, d):
        for permutation in self.all_perms(app.versions):
            appDirRoot = os.path.join(d, app.name + self.versionSuffix(permutation))
            if os.path.isdir(appDirRoot) or len(glob(appDirRoot + ".*")) > 0:
                return True
        return False

    def getVersionListsTopDir(self, fileName):
        # Show the framework how to find the version list given a file name
        # If it doesn't match, return None
        parts = os.path.basename(fileName).split(".")
        if len(parts) > 2 and parts[0] != "static_gui":
            # drop the run descriptor at the start and the date/time and pid at the end
            versionParts = ".".join(parts[1:-2]).split("++")
            return [part.split(".") for part in versionParts]

    def getVersionListSubDir(self, fileName, stem):
        # Show the framework how to find the version list given a file name
        # If it doesn't match, return None
        parts = fileName.split(".")
        if stem == parts[0]:
            # drop the application at the start
            return parts[1:]

    def getVersionSetSubDir(self, fileName, stem):
        vlist = self.getVersionListSubDir(fileName, stem)
        if vlist is not None:
            return frozenset(vlist)

    def getAllVersionLists(self, app, givenExtras, groupDirs):
        vlists = []
        for groupDir in groupDirs:
            for path in os.listdir(groupDir):
                fullPath = os.path.join(groupDir, path)
                if os.path.isdir(fullPath):
                    vlist = self.getVersionListSubDir(path, app.name)
                    if vlist is None:
                        continue
                    self.diag.info("Found list " + repr(vlist))
                    if givenExtras:
                        vset = frozenset(vlist).difference(givenExtras)
                        vlist = [v for v in vlist if v in vset]
                    if vlist not in vlists:
                        vlists.append(vlist)
        return vlists

    def expandExtraVersions(self, extras):
        expanded = set()
        for extra in extras:
            expanded.add(extra)
            expanded.update(extra.split("."))
        return expanded

    def getVersionsFromDirs(self, app, dirs, givenExtras):
        versions = []
        allGivenExtras = self.expandExtraVersions(givenExtras)
        self.diag.info("Getting extra versions from directories, versions from config = "
                       + repr(sorted(allGivenExtras)))
        appVersions = frozenset(app.versions)
        for versionLists, groupDirIter in groupby(dirs, self.getVersionListsTopDir):
            groupDirs = list(groupDirIter)
            self.diag.info("Considering version lists " + repr(versionLists) + " with dirs " + repr(groupDirs))
            for versionList in self.getAllVersionLists(app, allGivenExtras, groupDirs):
                version = ".".join(versionList)
                self.diag.info("Considering version list " + repr(versionList))
                versionSet = frozenset(versionList)
                if len(appVersions.difference(versionSet)) > 0:
                    continue  # If the given version isn't included, ignore it
                extraVersionSet = versionSet.difference(appVersions)
                # Important to preserve the order of the versions as received
                extraVersionList = [v for v in versionList if v in extraVersionSet]
                extraVersion = ".".join(extraVersionList)
                if len(groupDirs) == 1:
                    if extraVersion:
                        versions.append(extraVersion)
                        self.cacheRunDir(app, groupDirs[0], version)
                    else:
                        self.cacheRunDir(app, groupDirs[0])
                else:
                    datedVersionMap = {}
                    for dir in groupDirs:
                        datedVersionMap[os.path.basename(dir).split(".")[-2]] = dir
                    datedVersions = sorted(list(datedVersionMap.keys()), key=self.dateValue, reverse=True)
                    self.datedVersions.update(datedVersions)
                    self.diag.info("Found candidate dated versions: " + repr(datedVersions))
                    if not extraVersion:  # one of them has to be the main version...
                        mainVersion = datedVersions.pop(0)
                        self.cacheRunDir(app, datedVersionMap.get(mainVersion))
                    for datedVersion in datedVersions:
                        dir = datedVersionMap.get(datedVersion)
                        if extraVersion:
                            versions.append(extraVersion + "." + datedVersion)
                            self.cacheRunDir(app, dir, version + "." + datedVersion)
                        else:
                            versions.append(datedVersion)
                            if version:
                                self.cacheRunDir(app, dir, version + "." + datedVersion)
                            else:
                                self.cacheRunDir(app, dir, datedVersion)
        self.diag.info("Extra versions found as " + repr(versions))
        return versions

    @staticmethod
    def dateValue(version):
        yearlessDatetime = datetime.datetime.strptime(version, "%d%b%H%M%S")
        now = datetime.datetime.now()
        currYearDatetime = yearlessDatetime.replace(year=now.year)
        if currYearDatetime > now:
            return currYearDatetime.replace(year=now.year - 1)
        else:
            return currYearDatetime

    def checkSanity(self, app):
        if self.errorMessage:  # We failed already, basically
            raise plugins.TextTestError(self.errorMessage)

        runDir = self.findRunDir(app)
        if not runDir:
            raise plugins.TextTestWarning("Could not find any runs matching " + app.description())
        self.diag.info("Found run directory " + repr(runDir))
        self.reconnDir = self.findAppDirUnder(app, runDir)
        self.diag.info("Found application directory " + repr(self.reconnDir))
        if not self.reconnDir:
            raise plugins.TextTestWarning("Could not find an application directory matching " + app.description() +
                                          " for the run directory found at " + runDir)
        for datedVersion in self.datedVersions:
            app.addConfigEntry("unsaveable_version", datedVersion)


class ReconnectFilter(plugins.TextFilter):
    def __init__(self, rootDir):
        self.rootDir = rootDir

    def acceptsTestCase(self, test):
        return os.path.exists(os.path.join(self.rootDir, test.getRelPath()))

    def acceptsTestSuite(self, suite):
        return os.path.exists(os.path.join(self.rootDir, suite.getRelPath()))


class ReconnectTest(plugins.Action):
    def __init__(self, rootDirToCopy, fullRecalculate):
        self.rootDirToCopy = rootDirToCopy
        self.fullRecalculate = fullRecalculate
        self.diag = logging.getLogger("Reconnection")

    def __repr__(self):
        return "Reconnecting to"

    def __call__(self, test):
        newState = self.getReconnectState(test)
        self.describe(test, self.getStateText(newState))
        if newState:
            test.changeState(newState)

    def getReconnectState(self, test):
        reconnLocation = os.path.join(self.rootDirToCopy, test.getRelPath())
        self.diag.info("Reconnecting to test at " + reconnLocation)
        if os.path.isdir(reconnLocation):
            return self.getReconnectStateFrom(test, reconnLocation)
        else:
            return plugins.Unrunnable(briefText="no results",
                                      freeText="No file found to load results from under " + reconnLocation)

    def getStateText(self, state):
        if state:
            return " (state " + state.category + ")"
        else:
            return " (recomputing)"

    def getReconnectStateFrom(self, test, location, copyEvenIfLoadFails=True):
        stateToUse = None
        stateFile = os.path.join(location, "framework_tmp", "teststate")
        if os.path.isfile(stateFile):
            newTmpPath = os.path.dirname(self.rootDirToCopy)
            loaded, newState = test.getNewState(open(stateFile, "rb"), updatePaths=True, newTmpPath=newTmpPath)
            self.diag.info("Loaded state file at " + stateFile + " - " + repr(loaded))
            if loaded and self.modifyState(test, newState):  # if we can't read it, recompute it
                stateToUse = newState

        if (copyEvenIfLoadFails or stateToUse) and (self.fullRecalculate or not stateToUse):
            self.copyFiles(test, location)

        return stateToUse

    def copyFiles(self, test, reconnLocation):
        test.makeWriteDirectory()
        tmpDir = test.getDirectory(temporary=1)
        plugins.ensureDirectoryExists(tmpDir)
        self.diag.info("Copying files from " + reconnLocation + " to " + tmpDir)
        for file in os.listdir(reconnLocation):
            fullPath = os.path.join(reconnLocation, file)
            if os.path.isfile(fullPath):
                targetPath = os.path.join(tmpDir, os.path.basename(fullPath))
                try:
                    shutil.copyfile(fullPath, targetPath)
                except EnvironmentError as e:
                    # File could not be copied, may not have been readable
                    # Write the exception to it instead
                    targetFile = open(targetPath, "w")
                    targetFile.write("Failed to copy file - exception info follows :\n" + str(e) + "\n")
                    targetFile.close()

    def modifyState(self, test, newState):
        if self.fullRecalculate:
            # Only pick up errors here, recalculate the rest. Don't notify until
            # we're done with recalculation.
            if newState.hasResults():
                # Also pick up execution machines, we can't get them otherwise...
                test.state.executionHosts = newState.executionHosts
                return False  # don't actually change the state
            else:
                newState.lifecycleChange = ""  # otherwise it's regarded as complete
                return True
        else:
            return True

    def setUpApplication(self, app):
        plugins.log.info("Reconnecting to test results in directory " + self.rootDirToCopy)

    def setUpSuite(self, suite):
        self.describe(suite)




"""
Miscellaneous actions for generally housekeeping the state of the GUI
"""
from gi.repository import Gtk, GObject
from .. import guiplugins
from collections import OrderedDict
from texttestlib.default.batch import BatchApplicationData, MailSender
from texttestlib import plugins
import os
from texttestlib.jobprocess import killProcessAndChildren


class Quit(guiplugins.BasicActionGUI):
    def __init__(self, allApps, dynamic, inputOptions):
        guiplugins.BasicActionGUI.__init__(self, allApps, dynamic, inputOptions)
        self.runName = inputOptions.get("name", "") if dynamic else None

    def _getStockId(self):
        return "quit"

    def _getTitle(self):
        return "_Quit"

    def isActiveOnCurrent(self, *args):
        return True

    def getSignalsSent(self):
        return ["Quit"]

    def performOnCurrent(self):
        self.notify("Quit")

    def notifySetRunName(self, runName):
        if self.runName is not None:
            self.runName = runName

    def messageAfterPerform(self):
        pass  # GUI isn't there to show it

    def getConfirmationMessage(self):
        message = ""
        if self.runName and not self.runName.startswith("Tests started from"):
            message = "You named this run as follows : \n" + self.runName + "\n"
        runningProcesses = guiplugins.processMonitor.listQueryKillProcesses()
        if len(runningProcesses) > 0:
            message += "\nThese processes are still running, and will be terminated when quitting: \n\n   + " + \
                       "\n   + ".join(runningProcesses) + "\n"
        if message:
            message += "\nQuit anyway?\n"
        return message

    def notifyWindowClosed(self, *args):
        confirmationMessage = self.getConfirmationMessage()
        if confirmationMessage:
            dialog = self.showQueryDialog(self.getParentWindow(), confirmationMessage,
                                          Gtk.STOCK_DIALOG_WARNING, "Confirmation", None)
            responseId = dialog.run()
            saidCancel = responseId not in [Gtk.ResponseType.ACCEPT, Gtk.ResponseType.YES, Gtk.ResponseType.OK]
            dialog.hide()
            dialog.destroy()
            if saidCancel:
                self.cancel()
                raise guiplugins.CloseWindowCancelException("Closing window")
        self.notify("Quit")


class ResetGroups(guiplugins.BasicActionGUI):
    def isActiveOnCurrent(self, *args):
        return True

    def _getStockId(self):
        return "revert-to-saved"

    def _getTitle(self):
        return "R_eset"

    def messageAfterPerform(self):
        return "All options reset to default values."

    def getTooltip(self):
        return "Reset running options"

    def getSignalsSent(self):
        return ["Reset"]

    def performOnCurrent(self):
        self.notify("Reset")


class SetRunName(guiplugins.ActionDialogGUI):
    def __init__(self, *args):
        guiplugins.ActionDialogGUI.__init__(self, *args)
        self.addOption("name", "\nNew name for this run")

    def isActiveOnCurrent(self, *args):
        return True

    def _getStockId(self):
        return "index"

    def _getTitle(self):
        return "Set Run Name"

    def messageAfterPerform(self):
        pass

    def getDialogTitle(self):
        return "Set a new name for this run"

    def getTooltip(self):
        return "Provide a name for this run and warn before closing it"

    def getSignalsSent(self):
        return ["SetRunName"]

    def performOnCurrent(self):
        name = self.optionGroup.getOptionValue("name")
        self.notify("SetRunName", name)
        self.notify("Status", "Set name of run to '" + name + "'")


class RefreshAll(guiplugins.BasicActionGUI):
    def __init__(self, *args):
        guiplugins.BasicActionGUI.__init__(self, *args)
        self.rootTestSuites = []

    def _getTitle(self):
        return "Refresh"

    def _getStockId(self):
        return "refresh"

    def getTooltip(self):
        return "Refresh the whole test suite so that it reflects file changes"

    def messageBeforePerform(self):
        return "Refreshing the whole test suite..."

    def messageAfterPerform(self):
        return "Refreshed the test suite from the files"

    def addSuites(self, suites):
        self.rootTestSuites += suites

    def notifyRefresh(self):
        # when done indirectly
        self.performOnCurrent()

    def performOnCurrent(self):
        for suite in self.rootTestSuites:
            self.notify("ActionProgress")
            suite.app.setUpConfiguration()
            self.notify("ActionProgress")
            filters = suite.app.getFilterList(self.rootTestSuites)
            suite.refresh(filters)
            suite.refreshFilesRecursively()


class ViewScreenshots(guiplugins.ActionGUI):
    def _getTitle(self):
        return "View screenshots"

    def isActiveOnCurrent(self, *args):
        if len(self.currTestSelection) != 1:
            return False

        screenshotDir = self.getScreenshotDir()
        return os.path.isdir(screenshotDir)

    def getScreenshotDir(self):
        return os.path.join(self.currTestSelection[0].getDirectory(temporary=True), "screenshots")

    def performOnCurrent(self):
        screenshotDir = self.getScreenshotDir()
        allFiles = os.listdir(screenshotDir)
        allFiles.sort(key=self.getSortKey)
        allPaths = [os.path.join(screenshotDir, f) for f in allFiles]
        guiplugins.openLinkInBrowser(*allPaths)

    def getSortKey(self, fileName):
        number = fileName[10:-4]
        return int(number) if number.isdigit() else 0


class GenerateTestSummary(guiplugins.ActionDialogGUI):
    def __init__(self, *args):
        guiplugins.ActionDialogGUI.__init__(self, *args)
        self.addOption("generate", "", possibleDirs=[os.getenv("TEXTTEST_TMP", "")], saveFile=True)
        self.batchAppData = OrderedDict()
        self.allApps = OrderedDict()

    def performOnCurrent(self):
        fileName = self.getFileName()
        for test in self.currTestSelection:
            if test.state.isComplete():
                if test.app not in self.batchAppData:
                    self.addApplication(test)
                self.batchAppData[test.app].storeCategory(test)
        self.writeTextSummary(fileName)

    def writeTextSummary(self, fileName):
        mailSender = MailSender()
        with open(fileName, "w") as f:
            for appList in list(self.allApps.values()):
                batchDataList = list(map(self.batchAppData.get, appList))
                f.write(mailSender.makeContents(batchDataList, False))

    def getFileName(self):
        fileName = self.optionGroup.getOptionValue("generate")
        if not fileName:
            raise plugins.TextTestError("Cannot save selection - no file name specified")
        elif os.path.isdir(fileName):
            raise plugins.TextTestError("Cannot save selection - existing directory specified")
        else:
            return fileName

    def _getTitle(self):
        return "Generate test summary"

    def getRootSuite(self, test):
        if test.parent:
            return self.getRootSuite(test.parent)
        else:
            return test

    def addApplication(self, test):
        rootSuite = self.getRootSuite(test)
        app = test.app
        self.batchAppData[app] = BatchApplicationData(rootSuite)
        self.allApps.setdefault(app.name, []).append(app)


class ShowProcesses(guiplugins.ActionResultDialogGUI):
    def __init__(self, *args, **kw):
        guiplugins.ActionResultDialogGUI.__init__(self, *args, **kw)
        self.treeView = None

    def addContents(self):
        runningProcesses = guiplugins.processMonitor.getProcesses()
        if len(runningProcesses) > 0:
            processBox = self.createProcessBox(runningProcesses)
            self.dialog.vbox.pack_start(processBox, True, True, 0)
        else:
            messageBox = self.createDialogMessage(
                "No external processes have been launched from this TextTest instance.", Gtk.STOCK_DIALOG_INFO)
            self.dialog.vbox.pack_start(messageBox, True, True, 0)

    def makePopup(self):
        menu = Gtk.Menu()
        menuItem = Gtk.MenuItem("Kill Process")
        menu.append(menuItem)
        menuItem.connect("activate", self.killProcess)
        menuItem.show()
        return menu

    def killProcess(self, *args):
        iters = []

        def addSelIter(model, path, iter):
            proc = model.get_value(iter, 0)
            killProcessAndChildren(proc.pid)
            iters.append(iter)

        self.treeView.get_selection().selected_foreach(addSelIter)
        for iter in iters:
            self.treeView.get_model().remove(iter)

    def notifyProcessExited(self, pid):
        if self.treeView is None:
            return
        model = self.treeView.get_model()

        def removeIfPidMatches(m, path, iter):
            proc = model.get_value(iter, 0)
            if proc.pid == pid:
                model.remove(iter)
                return True
        model.foreach(removeIfPidMatches)

    def showPopup(self, treeview, event, popupMenu):
        if event.button == 3:
            pathInfo = treeview.get_path_at_pos(int(event.x), int(event.y))
            if pathInfo is not None:
                treeview.grab_focus()
                popupMenu.popup(None, None, None, event.button, event.time)
                treeview.stop_emission("button_press_event")  # Disable default handler which auto-selects rows

    def setPid(self, column, cell, model, iter):
        cell.set_property('text', str(model.get_value(iter, 0).pid))

    def createProcessBox(self, runningProcesses):
        listStore = Gtk.ListStore(GObject.TYPE_PYOBJECT, str)
        for proc, description in runningProcesses:
            listStore.append([proc, description])
        self.treeView = Gtk.TreeView(listStore)
        self.treeView.set_name("Process Tree View")

        cell = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("PID")
        column.pack_start(cell, True)
        column.set_cell_data_func(cell, self.setPid)

        self.treeView.append_column(column)
        cell2 = Gtk.CellRendererText()
        column = Gtk.TreeViewColumn("Description", cell2, text=1)
        self.treeView.append_column(column)

        self.treeView.get_selection().set_mode(Gtk.SelectionMode.MULTIPLE)
        popup = self.makePopup()
        self.treeView.connect("button_press_event", self.showPopup, popup)
        return self.treeView

    def isActiveOnCurrent(self, *args):
        return True

    def _getTitle(self):
        return "Show Processes"


def getInteractiveActionClasses(dynamic):
    classes = [Quit, SetRunName, ShowProcesses]
    if dynamic:
        classes.append(ViewScreenshots)
        classes.append(GenerateTestSummary)
    else:
        classes += [RefreshAll, ResetGroups]
    return classes




import datetime
import time
import os
from . import vcs_independent


class BzrInterface(vcs_independent.VersionControlInterface):
    def __init__(self, controlDir):
        warningStates = ["Modified", "Removed", "Added", "Renamed"]
        errorStates = ["Unknown", "Conflicts", "Kind changed"]
        vcs_independent.VersionControlInterface.__init__(self, controlDir, "Bazaar", warningStates, errorStates, "-1")
        self.defaultArgs["rm"] = ["--force"]

    def isVersionControlled(self, dirname):
        args = self.getCmdArgs("status") + [dirname]
        returncode, output, err = self.getProcessResults(args)
        # Unless the result is "unknown:" followed by the relpath of the dirname, it's version controlled, at least a bit...
        if returncode and err:
            return False
        lines = output.splitlines()
        if len(lines) != 2:
            return True
        if lines[0].strip() != "unknown:":
            return True
        pathname = os.path.normpath(lines[1].strip())
        return not dirname.endswith(pathname)

    def getDateFromLog(self, output):
        for line in output.splitlines():
            if line.startswith("timestamp:"):
                dateStr = " ".join(line.split()[2:4])
                return datetime.datetime(*(self.parseDateTime(dateStr)[0:6]))

    def getGraphicalDiffArgs(self, diffProgram):
        return ["bzr", "diff", "--using=" + diffProgram]

    def parseDateTime(self, input):
        return time.strptime(input, "%Y-%m-%d %H:%M:%S")

    def getStateFromStatus(self, output):
        for line in reversed(output.splitlines()):
            if line.endswith(":"):
                return line[:-1].capitalize()
        return "Unchanged"

    def getCombinedRevisionOptions(self, r1, r2):
        return ["-r", r1 + ".." + r2]

    def removePath(self, path):
        retCode = self.callProgram("rm", [path])  # Always removes
        return retCode == 0

    # Hack for bug in Bazaar, which can't handle symbolic links to the branch...
    def callProgramOnFiles(self, cmdName, fileArg, recursive=False, extraArgs=[], **kwargs):
        if cmdName == "add":
            basicArgs = self.getCmdArgs(cmdName, extraArgs)
            for fileName in self.getFileNamesForCmd(cmdName, fileArg, recursive):
                self.callProgramWithHandler(fileName, basicArgs + [os.path.realpath(fileName)], **kwargs)
        else:
            vcs_independent.VersionControlInterface.callProgramOnFiles(
                self, cmdName, fileArg, recursive, extraArgs, **kwargs)


vcs_independent.vcsClass = BzrInterface
InteractiveActionConfig = vcs_independent.InteractiveActionConfig



from gi.repository import Gtk, Gdk
import os
import time
import stat
from texttestlib import plugins
from .. import guiplugins
from collections import OrderedDict

# pwd and grp doesn't exist on windows ...
try:
    import pwd
    import grp
except ImportError:
    pass


class FileProperties:
    def __init__(self, path):
        self.abspath = path
        self.filename = os.path.basename(self.abspath)
        self.dir = os.path.dirname(self.abspath)
        self.status = os.lstat(self.abspath)
        self.now = int(time.time())
        self.recent = self.now - (6 * 30 * 24 * 60 * 60)  # 6 months ago

    def inqType(self):
        mode = self.status[stat.ST_MODE]
        if stat.S_ISLNK(mode):
            return "l"
        elif stat.S_ISDIR(mode):
            return "d"
        else:
            return "-"

    def inqMode(self):
        permissions = ""
        for who in "USR", "GRP", "OTH":
            for what in "R", "W", "X":
                # lookup attribute at runtime using getattr
                if self.status[stat.ST_MODE] & getattr(stat, "S_I" + what + who):
                    permissions = permissions + what.lower()
                else:
                    permissions = permissions + "-"
        return permissions

    def inqLinks(self):
        return self.status[stat.ST_NLINK]

    def inqOwner(self):
        try:
            uid = self.status[stat.ST_UID]
            return str(pwd.getpwuid(uid)[0])
        except Exception:  # KeyError, AttributeError (on Windows) possible
            return "?"

    def inqGroup(self):
        try:
            gid = self.status[stat.ST_GID]
            return str(grp.getgrgid(gid)[0])
        except Exception:  # KeyError, AttributeError (on Windows) possible
            return "?"

    def inqSize(self):
        return self.status[stat.ST_SIZE]

    def formatTime(self, timeStamp):
        # %e is more appropriate than %d below, as it fills with space
        # rather than 0, but it is not supported on Windows, it seems.
        if timeStamp < self.recent or timeStamp > self.now:
            timeFormat = "%b %d  %Y"
        else:
            timeFormat = "%b %d %H:%M"
        return time.strftime(timeFormat, time.localtime(timeStamp))

    def inqModificationTime(self):
        return self.formatTime(self.status[stat.ST_MTIME])

    # Return the *nix type format:
    # -rwxr--r--    1 mattias carm       1675 Nov 16  1998 .xinitrc_old
    def getUnixRepresentation(self):
        return (self.inqType(), self.inqMode(),
                self.inqLinks(), self.inqOwner(),
                self.inqGroup(), self.inqSize(),
                self.inqModificationTime(), self.filename)


class ShowFileProperties(guiplugins.ActionResultDialogGUI):
    def __init__(self, allApps, dynamic, *args):
        self.dynamic = dynamic
        guiplugins.ActionResultDialogGUI.__init__(self, allApps)

    def _getStockId(self):
        return "properties"

    def isActiveOnCurrent(self, *args):
        return ((not self.dynamic) or len(self.currTestSelection) == 1) and \
            len(self.currFileSelection) > 0

    def _getTitle(self):
        return "_File Properties"

    def getTooltip(self):
        return "Show properties of selected files"

    def describeTests(self):
        return str(len(self.currFileSelection)) + " files"

    def getAllProperties(self):
        errors, properties = [], []
        for file, comp in self.currFileSelection:
            if self.dynamic and comp:
                self.processFile(comp.tmpFile, properties, errors)
                self.processFile(comp.stdFile, properties, errors)
            else:
                self.processFile(file, properties, errors)

        if len(errors):
            raise plugins.TextTestError("Failed to get file properties:\n" + "\n".join(errors))

        return properties

    def processFile(self, file, properties, errors):
        if file:
            try:
                prop = FileProperties(file)
                properties.append(prop)
            except Exception as e:
                errors.append(str(e))

    # xalign = 1.0 means right aligned, 0.0 means left aligned
    def justify(self, text, xalign=0.0):
        alignment = Gtk.Alignment.new(xalign, 0.0, 0.0, 0.0)
        label = Gtk.Label(label=text)
        alignment.add(label)
        return alignment

    def addContents(self):
        dirToProperties = OrderedDict()
        props = self.getAllProperties()
        for prop in props:
            dirToProperties.setdefault(prop.dir, []).append(prop)
        vbox = self.createVBox(dirToProperties)
        self.dialog.vbox.pack_start(vbox, True, True, 0)

    def createVBox(self, dirToProperties):
        vbox = Gtk.VBox()
        for dir, properties in list(dirToProperties.items()):
            expander = Gtk.Expander()
            expander.set_label_widget(self.justify(dir))
            table = Gtk.Table(len(properties), 7)
            table.set_col_spacings(5)
            row = 0
            for prop in properties:
                values = prop.getUnixRepresentation()
                table.attach(self.justify(values[0] + values[1], 1.0), 0, 1, row, row + 1)
                table.attach(self.justify(values[2], 1.0), 1, 2, row, row + 1)
                table.attach(self.justify(values[3], 0.0), 2, 3, row, row + 1)
                table.attach(self.justify(values[4], 0.0), 3, 4, row, row + 1)
                table.attach(self.justify(values[5], 1.0), 4, 5, row, row + 1)
                table.attach(self.justify(values[6], 1.0), 5, 6, row, row + 1)
                table.attach(self.justify(prop.filename, 0.0), 6, 7, row, row + 1)
                row += 1
            hbox = Gtk.HBox()
            hbox.pack_start(table, False, False, 0)
            innerBorder = Gtk.Alignment.new(0.0, 0.0, 0.0, 0.0)
            innerBorder.set_padding(5, 0, 0, 0)
            innerBorder.add(hbox)
            expander.add(innerBorder)
            expander.set_expanded(True)
            border = Gtk.Alignment.new(0.0, 0.0, 0.0, 0.0)
            border.set_padding(5, 5, 5, 5)
            border.add(expander)
            vbox.pack_start(border, False, False, 0)
        return vbox


class CopyPathToClipboard(guiplugins.ActionGUI):
    def _getTitle(self):
        return "Copy Path To Clipboard"

    def messageAfterPerform(self):
        return "Copied full path of selected file to clipboard."

    def isActiveOnCurrent(self, *args):
        return len(self.currFileSelection) == 1

    def performOnCurrent(self):
        fileName, comp = self.currFileSelection[0]
        if comp and hasattr(comp, "tmpFile"):
            fileName = comp.tmpFile
        # Copy to both, for good measure, avoid problems with e.g. Exceed configuration
        for clipboard in [Gtk.Clipboard.get(Gdk.SELECTION_CLIPBOARD), Gtk.Clipboard.get(Gdk.SELECTION_PRIMARY)]:
            clipboard.set_text(fileName, -1)


def getInteractiveActionClasses():
    return [ShowFileProperties, CopyPathToClipboard]




"""
The various classes that launch external programs to view files
"""

import os
from texttestlib import plugins
from .. import guiplugins
from string import Template
from copy import copy
from tempfile import mkstemp
import subprocess


class FileViewAction(guiplugins.ActionGUI):
    def __init__(self, *args, **kw):
        self.performArgs = []
        guiplugins.ActionGUI.__init__(self, *args, **kw)

    def singleTestOnly(self):
        return True

    def isActiveOnCurrent(self, *args):
        if not guiplugins.ActionGUI.isActiveOnCurrent(self):
            return False
        for fileName, obj in self.currFileSelection:
            if self.isActiveForFile(fileName, obj):
                return True
        return False

    def isActiveForFile(self, fileName, *args):
        return fileName and not os.path.isdir(fileName)

    def useFiltered(self):
        return False

    def getLargestFileSize(self, f, *args):
        try:
            return os.path.getsize(f)
        except EnvironmentError:
            return 0

    def getConfirmationMessage(self):
        self.performArgs = []
        message = ""
        for fileName, associatedObject in self.currFileSelection:
            if self.isActiveForFile(fileName, associatedObject):
                message, args = self.getConfMessageForFile(fileName, associatedObject)
                self.performArgs.append(args)
        return message

    def performOnCurrent(self):
        for args in self.performArgs:
            self.performOnFile(*args)

    def performOnFile(self, viewTool, *args):
        try:
            self._performOnFile(viewTool, *args)
        except OSError:
            self.showErrorDialog("Cannot find " + self.getToolDescription() + " '" + viewTool +
                                 "'.\nPlease install it somewhere on your PATH or\n"
                                 "change the configuration entry '" + self.getToolConfigEntry() + "'.")

    def getConfMessageForFile(self, fileName, associatedObject):
        fileToView = self.getFileToView(fileName, associatedObject)
        if os.path.isfile(fileToView) or os.path.islink(fileToView):
            viewTool = self.getViewToolName(fileToView)
            if viewTool:
                args = (viewTool, fileToView, associatedObject)
                maxFileSize = plugins.parseBytes(self.getConfigValue("max_file_size", viewTool))
                if maxFileSize >= 0:
                    largestFileSize = self.getLargestFileSize(fileToView, associatedObject)
                    if largestFileSize > maxFileSize:
                        message = "You are trying to view a file of size " + str(largestFileSize) + " bytes, while a limit of " + \
                                  str(maxFileSize) + " bytes is set for the tool '" + \
                            viewTool + "'. Are you sure you wish to continue?"
                        return message, args

                return "", args
            else:
                raise plugins.TextTestError("No " + self.getToolDescription() + " is defined for files of type '" +
                                            os.path.basename(fileToView).split(".")[0] +
                                            "'.\nPlease point the configuration entry '" + self.getToolConfigEntry() +
                                            "' at a valid program to view the file.")
        else:
            raise plugins.TextTestError("File '" + os.path.basename(fileName) +
                                        "' cannot be viewed as it has been removed in the file system." + self.noFileAdvice())

    def isDefaultViewer(self, *args):
        return False

    def extraPostfix(self):
        return ""

    def notifyViewFile(self, fileName, comp, newFile):
        if self.isDefaultViewer(comp, newFile):
            allArgs = (fileName, comp)
            self.currFileSelection = [allArgs]
            self.runInteractive()

    def getFileToView(self, fileName, associatedObject):
        try:
            # associatedObject might be a comparison object, but it might not
            # Use the comparison if it's there
            return associatedObject.existingFile(self.useFiltered(), self.extraPostfix())
        except AttributeError:
            return fileName

    def noFileAdvice(self):
        if len(self.currAppSelection) > 0:
            return "\n" + self.currAppSelection[0].noFileAdvice()
        else:
            return ""

    def testDescription(self):
        if len(self.currTestSelection) > 0:
            return " (from test " + self.currTestSelection[0].uniqueName + ")"
        else:
            return ""

    def getRemoteHost(self):
        if len(self.currTestSelection) > 0:
            state = self.currTestSelection[0].stateInGui
            if hasattr(state, "executionHosts") and len(state.executionHosts) > 0:
                return plugins.interpretHostname(state.executionHosts[0])
        return "localhost"

    def getRemoteArgs(self, cmdArgs):
        remoteHost = self.getRemoteHost()
        return self.currTestSelection[0].app.getCommandArgsOn(remoteHost, cmdArgs, graphical=True)

    def getSignalsSent(self):
        return ["ViewerStarted"]

    def startViewer(self, cmdArgs, description, checkOutput=False, **kwargs):
        testDesc = self.testDescription()
        fullDesc = description + testDesc
        nullFile = open(os.devnull, "w")
        stdout = subprocess.PIPE if checkOutput else nullFile
        self.notify("Status", 'Started "' + description + '" in background' + testDesc + '.')
        guiplugins.processMonitor.startProcess(cmdArgs, fullDesc, stdout=stdout, stderr=nullFile, **kwargs)
        self.notifyThreaded("ViewerStarted")  # Don't call application events directly in the GUI thread

    def getStem(self, fileName):
        return os.path.basename(fileName).split(".")[0]

    def testRunning(self):
        if len(self.currTestSelection) > 0:
            return self.currTestSelection[0].stateInGui.hasStarted() and \
                not self.currTestSelection[0].stateInGui.isComplete()
        else:
            return False

    def getViewToolName(self, fileName):
        stem = self.getStem(fileName)
        return self.getConfigValue(self.getToolConfigEntry(), stem)

    def getConfigValue(self, *args):
        if len(self.currTestSelection) > 0:
            return self.currTestSelection[0].getCompositeConfigValue(*args)
        else:
            return guiplugins.guiConfig.getCompositeValue(*args)

    def differencesActive(self, comparison):
        if not comparison or comparison.newResult() or comparison.missingResult():
            return False
        return comparison.hasDifferences()

    def messageAfterPerform(self):
        pass  # provided by starting viewer, with message


class ViewInEditor(FileViewAction):
    def __init__(self, allApps, dynamic, *args):
        FileViewAction.__init__(self, allApps)
        self.dynamic = dynamic

    def _getStockId(self):
        return "open"

    def getToolConfigEntry(self):
        return "view_program"

    def getToolDescription(self):
        return "file viewing program"

    def viewFile(self, fileName, viewTool, exitHandler, exitHandlerArgs):
        cmdArgs, descriptor, env = self.getViewCommand(fileName, viewTool)
        description = descriptor + " " + os.path.basename(fileName)
        checkOutput = cmdArgs[0] == "storytext_editor"
        self.startViewer(cmdArgs, description=description, checkOutput=checkOutput, env=env,
                         exitHandler=exitHandler, exitHandlerArgs=exitHandlerArgs)

    def getViewerEnvironment(self, cmdArgs):
        # An absolute path to the viewer may indicate a custom tool, send the test environment along too
        # Doing this is unlikely to cause harm in any case
        if len(self.currTestSelection) > 0:
            if cmdArgs[0] == "storytext_editor":
                storytextHome = self.currTestSelection[0].getEnvironment("STORYTEXT_HOME")
                return plugins.copyEnvironment({"STORYTEXT_HOME": storytextHome}, ignoreVars=["USECASE_RECORD_SCRIPT", "USECASE_REPLAY_SCRIPT"])
            elif os.path.isabs(cmdArgs[0]):
                return self.currTestSelection[0].getRunEnvironment()

    def getViewCommand(self, fileName, viewProgram):
        # viewProgram might have arguments baked into it...
        cmdArgs = plugins.splitcmd(viewProgram) + [fileName]
        program = cmdArgs[0]
        descriptor = " ".join([os.path.basename(program)] + cmdArgs[1:-1])
        env = self.getViewerEnvironment(cmdArgs)
        interpreter = plugins.getInterpreter(program)
        if interpreter:
            cmdArgs = [interpreter] + cmdArgs

        if guiplugins.guiConfig.getCompositeValue("view_file_on_remote_machine", self.getStem(fileName)):
            cmdArgs = self.getRemoteArgs(cmdArgs)

        return cmdArgs, descriptor, env

    def _performOnFile(self, viewTool, fileName, *args):
        exitHandler, exitHandlerArgs = self.findExitHandlerInfo(fileName, *args)
        return self.viewFile(fileName, viewTool, exitHandler, exitHandlerArgs)

    def editingComplete(self):
        self.applicationEvent("file editing operations to complete")


class ViewConfigFileInEditor(ViewInEditor):
    def __init__(self, *args):
        ViewInEditor.__init__(self, *args)
        self.rootTestSuites = []

    def _getTitle(self):
        return "View In Editor"

    def addSuites(self, suites):
        self.rootTestSuites += suites

    def isActiveOnCurrent(self, *args):
        return len(self.currFileSelection) > 0

    def notifyViewApplicationFile(self, fileName, apps, *args):
        self.currFileSelection = [(fileName, apps)]
        self.runInteractive()

    def notifyViewReadonlyFile(self, fileName):
        viewTool = self.getConfigValue("view_program", "default")
        return self.viewFile(fileName, viewTool, self.applicationEvent, ("the readonly file viewer to be closed",))

    def findExitHandlerInfo(self, dummy, apps):
        return self.configFileChanged, (apps,)

    def getSignalsSent(self):
        return ["ReloadConfig"]

    def configFileChanged(self, apps):
        for app in apps:
            app.setUpConfiguration()
            suite = self.findSuite(app)
            suite.refreshFilesRecursively()

        # May have affected e.g. imported files, script references etc
        self.notify("ReloadConfig")
        self.editingComplete()

    def findSuite(self, app):
        for suite in self.rootTestSuites:
            if suite.app is app:
                return suite


class ViewTestFileInEditor(ViewInEditor):
    def _getTitle(self):
        return "View File"

    def isDefaultViewer(self, comparison, isNewFile):
        return not isNewFile and not self.differencesActive(comparison) and \
            (not self.testRunning() or not guiplugins.guiConfig.getValue("follow_file_by_default"))

    def findExitHandlerInfo(self, fileName, *args):
        if self.dynamic:
            return self.editingComplete, ()

        # options file can change appearance of test (environment refs etc.)
        baseName = os.path.basename(fileName)
        for stem in ["options", "testsuite", "config"]:
            if baseName.startswith(stem):
                tests = self.getTestsForFile(stem, fileName)
                if len(tests) > 0:
                    methodName = "handle" + stem.capitalize() + "Edit"
                    return getattr(self, methodName), (tests,)
        return self.staticGUIEditingComplete, (copy(self.currTestSelection), fileName)

    def getTestsForFile(self, stem, fileName):
        tests = []
        for test in self.currTestSelection:
            defFile = test.getFileName(stem)
            if defFile and plugins.samefile(fileName, defFile):
                tests.append(test)
        return tests

    def handleTestsuiteEdit(self, suites):
        for suite in suites:
            suite.refresh(suite.app.getFilterList(suites))
        self.editingComplete()

    def handleOptionsEdit(self, tests):
        for test in tests:
            test.filesChanged()
        self.editingComplete()

    def handleConfigEdit(self, tests):
        for test in tests:
            test.reloadTestConfigurations()
        self.editingComplete()

    def getSignalsSent(self):
        return ["RefreshFilePreviews"] + ViewInEditor.getSignalsSent(self)

    def staticGUIEditingComplete(self, tests, fileName):
        for test in tests:
            self.notify("RefreshFilePreviews", test, fileName)
        self.editingComplete()


class EditTestFileInEditor(ViewTestFileInEditor):
    def _getTitle(self):
        return "Edit File"

    def getViewToolName(self, *args):
        return self.getConfigValue(self.getToolConfigEntry(), "default")

    def isDefaultViewer(self, comp, isNewFile):
        return isNewFile

    def _getStockId(self):
        pass  # don't use same stock for both


class ViewFilteredTestFileInEditor(ViewTestFileInEditor):
    def _getStockId(self):
        pass  # don't use same stock for both

    def useFiltered(self):
        return True

    def _getTitle(self):
        return "View Filtered File"

    def isActiveForFile(self, fileName, comparison):
        return bool(comparison)

    def isDefaultViewer(self, *args):
        return False


class ContentFilterViewer:
    def extraPostfix(self):
        return ".normal"

    def unorderedFiltersActive(self, comparison):
        return len(self.currAppSelection[0].getCompositeConfigValue("unordered_text", comparison.stem)) > 0


class ViewContentFilteredTestFileInEditor(ContentFilterViewer, ViewFilteredTestFileInEditor):
    def _getTitle(self):
        return "View Content-Filtered File"

    def isActiveForFile(self, fileName, comparison):
        return ViewFilteredTestFileInEditor.isActiveForFile(self, fileName, comparison) and \
            self.unorderedFiltersActive(comparison)


class ViewFilteredOrigFileInEditor(ViewFilteredTestFileInEditor):
    def _getTitle(self):
        return "View Filtered Original File"

    def isActiveForFile(self, fileName, comparison):
        return comparison and not comparison.newResult()

    def getFileToView(self, fileName, comparison):
        return comparison.getStdFile(self.useFiltered(), self.extraPostfix())


class ViewOrigFileInEditor(ViewFilteredOrigFileInEditor):
    def _getTitle(self):
        return "View Original File"

    def useFiltered(self):
        return False


class EditOrigFileInEditor(ViewOrigFileInEditor):
    def _getTitle(self):
        return "Edit Original File"

    def getViewToolName(self, *args):
        return self.getConfigValue(self.getToolConfigEntry(), "default")


class ViewContentFilteredOrigFileInEditor(ContentFilterViewer, ViewFilteredOrigFileInEditor):
    def _getTitle(self):
        return "View Content-Filtered Original File"

    def isActiveForFile(self, fileName, comparison):
        return ViewFilteredOrigFileInEditor.isActiveForFile(self, fileName, comparison) and \
            self.unorderedFiltersActive(comparison)


class ViewFileDifferences(FileViewAction):
    def _getTitle(self):
        return "View Raw Differences"

    def getToolConfigEntry(self):
        return "diff_program"

    def getToolDescription(self):
        return "graphical difference program"

    def isActiveForFile(self, fileName, comparison):
        if bool(comparison):
            if not (comparison.newResult() or comparison.missingResult()):
                return True
        return False

    def getLargestFileSize(self, tmpFile, comparison):
        stdFile = comparison.getStdFile(self.useFiltered(), self.extraPostfix())
        return max(os.path.getsize(stdFile), os.path.getsize(tmpFile))

    def _performOnFile(self, diffProgram, tmpFile, comparison):
        stdFile = comparison.getStdFile(self.useFiltered(), self.extraPostfix())
        self.runDiff(diffProgram, stdFile, tmpFile)

    def runDiff(self, diffProgram, stdFile, tmpFile):
        description = diffProgram + " " + os.path.basename(stdFile) + " " + os.path.basename(tmpFile)
        cmdArgs = plugins.splitcmd(diffProgram) + [stdFile, tmpFile]
        self.startViewer(cmdArgs, description=description, exitHandler=self.diffingComplete)

    def diffingComplete(self, *args):
        self.applicationEvent("the " + self.getToolDescription() + " to terminate")


class PairwiseFileViewer:
    def singleTestOnly(self):
        return False

    def isActiveOnCurrent(self, *args):
        return len(self.currTestSelection) == 2 and len(self.currFileSelection) == 1 and \
            FileViewAction.isActiveOnCurrent(self, *args)

    def isActiveForFile(self, fileName, comparison):
        if bool(comparison) and not comparison.missingResult():
            otherComp = self.findOtherComparison(comparison)
            if otherComp and not otherComp.missingResult():
                return True
        return False

    def findOtherComparison(self, comparison):
        state = self.currTestSelection[-1].stateInGui
        return state.findComparison(comparison.stem, includeSuccess=True)[0] if hasattr(state, "findComparison") else None

    def _performOnFile(self, diffProgram, tmpFile, comparison):
        otherComp = self.findOtherComparison(comparison)
        file1 = comparison.getTmpFile(self.useFiltered(), self.extraPostfix())
        file2 = otherComp.getTmpFile(self.useFiltered(), self.extraPostfix())
        self.runDiff(diffProgram, file1, file2)


class ViewPairwiseDifferences(PairwiseFileViewer, ViewFileDifferences):
    def _getTitle(self):
        return "View Raw Differences (between tests)"


class ViewFilteredFileDifferences(ViewFileDifferences):
    def _getTitle(self):
        return "View Differences"

    def useFiltered(self):
        return True

    def isActiveForFile(self, fileName, comparison):
        return self.differencesActive(comparison)

    def isDefaultViewer(self, comparison, *args):
        return self.differencesActive(comparison)


class ViewFilteredPairwiseDifferences(PairwiseFileViewer, ViewFilteredFileDifferences):
    def _getTitle(self):
        return "View Differences (between tests)"

    def useFiltered(self):
        return True

    def isDefaultViewer(self, comparison, *args):
        return False


class ViewContentFilteredFileDifferences(ContentFilterViewer, ViewFilteredFileDifferences):
    def _getTitle(self):
        return "View Content-Filtered Differences"

    def isActiveForFile(self, fileName, comparison):
        return ViewFileDifferences.isActiveForFile(self, fileName, comparison) and self.unorderedFiltersActive(comparison)

    def isDefaultViewer(self, comparison, *args):
        return False


class ViewContentFilteredPairwiseDifferences(PairwiseFileViewer, ViewContentFilteredFileDifferences):
    def _getTitle(self):
        return "View Content-Filtered Differences (between tests)"


class FollowFile(FileViewAction):
    def _getTitle(self):
        return "Follow File Progress"

    def getToolConfigEntry(self):
        return "follow_program"

    def getToolDescription(self):
        return "file-following program"

    def isActiveForFile(self, *args):
        return self.testRunning()

    def fileToFollow(self, fileName, comparison):
        if comparison:
            return comparison.tmpFile
        else:
            return fileName

    def isDefaultViewer(self, comparison, *args):
        return not self.differencesActive(comparison) and self.testRunning() and \
            guiplugins.guiConfig.getValue("follow_file_by_default")

    def getFollowProgram(self, followProgram, fileName):
        title = '"File ' + os.path.basename(fileName) + " from test " + self.currTestSelection[0].name + '"'
        envDir = {"TEXTTEST_FOLLOW_FILE_TITLE": title}  # Title of the window when following file progress
        return Template(followProgram).safe_substitute(envDir)

    def getFollowCommand(self, program, fileName):
        localArgs = plugins.splitcmd(program) + [fileName]
        return self.getRemoteArgs(localArgs)

    def _performOnFile(self, followProgram, fileName, comparison):
        useFile = self.fileToFollow(fileName, comparison)
        useProgram = self.getFollowProgram(followProgram, fileName)
        program = useProgram.split()[0]
        description = useProgram + " " + os.path.basename(useFile)
        cmdArgs = self.getFollowCommand(useProgram, useFile)
        if self.canRunCommand(program):
            self.startViewer(cmdArgs, description=description, exitHandler=self.followComplete)
        else:
            self.showErrorDialog("Couldn't find " + program + " to run the " + self.getToolDescription())

    def followComplete(self, *args):
        self.applicationEvent("the " + self.getToolDescription() + " to terminate")

    def canRunCommand(self, cmd):
        if len(self.currTestSelection) > 0:
            app = self.currTestSelection[0].app
        if self.getRemoteHost() != "localhost":
            retcode = app.runCommandOn(self.getRemoteHost(), ["which", cmd], collectExitCode=True)
            return retcode == 0
        else:
            return True


class EditTestDescription(EditTestFileInEditor):
    def _getTitle(self):
        return "Edit Description"

    def isActiveOnCurrent(self, *args):
        if not guiplugins.ActionGUI.isActiveOnCurrent(self):
            return False
        return len(self.currTestSelection) == 1

    def performOnCurrent(self):
        test = self.currTestSelection[0]
        fileName = self.createTmpFile(test.description)
        _, args = self.getConfMessageForFile(fileName, test)
        self.performOnFile(*args)

    def createTmpFile(self, text):
        tmpFile, fileName = mkstemp(text=True)
        os.write(tmpFile, text.encode())
        os.close(tmpFile)
        return fileName

    def findExitHandlerInfo(self, fileName, test):
        return self.updateDescription, (fileName, test)

    def updateDescription(self, fileName, test):
        text = self.getEditedText(fileName)
        test.rename(test.name, text)
        if os.path.exists(fileName):
            os.remove(fileName)
        self.editingComplete()

    def getEditedText(self, fileName):
        text = ""
        with open(fileName, "r", encoding="utf8") as f:
            for line in f:
                text += line
        return text

    def isDefaultViewer(self, *args):
        return False


def getInteractiveActionClasses(dynamic):
    classes = [ViewTestFileInEditor, EditTestFileInEditor]
    if dynamic:
        classes += [ViewFilteredTestFileInEditor, ViewContentFilteredTestFileInEditor,
                    ViewOrigFileInEditor, EditOrigFileInEditor, ViewContentFilteredOrigFileInEditor, ViewFilteredOrigFileInEditor,
                    ViewFileDifferences, ViewContentFilteredFileDifferences, ViewFilteredFileDifferences,
                    ViewPairwiseDifferences, ViewContentFilteredPairwiseDifferences, ViewFilteredPairwiseDifferences,
                    FollowFile]
    else:
        classes.append(ViewConfigFileInEditor)
        classes.append(EditTestDescription)

    return classes



""" Code related to building the summary page and the graphs etc. """

from . import testoverview
import logging
import os
import shutil
import time
import operator
import sys
from texttestlib import plugins
from html.parser import HTMLParser
from collections import OrderedDict
from glob import glob
from .batchutils import BatchVersionFilter, parseFileName, convertToUrl, getEnvironmentFromRunFiles
import datetime
from functools import reduce


class GenerateFromSummaryData(plugins.ScriptWithArgs):
    locationApps = OrderedDict()
    summaryFileName = "index.html"
    basePath = ""

    def __init__(self, args=[""]):
        argDict = self.parseArguments(args, ["basepath", "file"])
        if "basepath" in argDict:
            GenerateFromSummaryData.basePath = argDict["basepath"]
        if "file" in argDict:
            GenerateFromSummaryData.summaryFileName = argDict["file"]

    def setUpApplication(self, app):
        location = os.path.realpath(app.getBatchConfigValue("historical_report_location"))
        usePie = app.getBatchConfigValue("historical_report_piechart_summary") == "true"
        versionFilter = BatchVersionFilter(app.getBatchSession())
        rejected = bool(versionFilter.findUnacceptableVersion(app))
        self.locationApps.setdefault(location, []).append((app, usePie, rejected))

    def generateForApps(self, apps):
        def shouldGenerate(currApps):
            for app in apps:
                for currApp, _, _ in currApps:
                    if app is currApp or app in currApp.extras:
                        return True
            return False
        return self.finalise(shouldGenerate)

    @classmethod
    def finalise(cls, predicate=None):
        for location, apps in list(cls.locationApps.items()):
            if predicate is None or predicate(apps):
                if not all((rejected for app, usePie, rejected in apps)):
                    defaultUsePie = all((usePie for app, usePie, rejected in apps))
                    plugins.log.info("Generating index page at " + os.path.join(location,
                                                                                cls.summaryFileName) + ", from following:")
                    for app, _, rejected in apps:
                        text = "- " + app.description()
                        if rejected:
                            text += " (rejected)"
                        plugins.log.info(text)
                    dataFinder = SummaryDataFinder(location, apps, cls.summaryFileName, cls.basePath, defaultUsePie)
                    appsWithVersions = dataFinder.getAppsWithVersions()
                    if appsWithVersions:
                        cls.generate(dataFinder, appsWithVersions, apps[0][0].getConfigValue("file_to_url"))
                else:
                    plugins.log.info("No applications generated for index page at " +
                                     repr(os.path.join(location, cls.summaryFileName)) + ".")


class GenerateSummaryPage(GenerateFromSummaryData):
    scriptDoc = "Generate a summary page which links all the other generated pages"

    @classmethod
    def generate(cls, *args):
        generator = SummaryGenerator()
        generator.generatePage(*args)


class GenerateGraphs(GenerateFromSummaryData):
    scriptDoc = "Generate standalone graphs along the lines of the ones that appear in the HTML report"

    @classmethod
    def generate(cls, dataFinder, appsWithVersions, *args):
        from .resultgraphs import GraphGenerator
        for appName, versions in sorted(list(appsWithVersions.items())):
            for version in versions:
                results = dataFinder.getAllSummaries(appName, version)
                if len(results) > 1:
                    fileName = dataFinder.getGraphFile(appName, version)
                    graphTitle = "Test results for Application: '" + appName + "'  Version: '" + version + "'"
                    graphGenerator = GraphGenerator()
                    graphGenerator.generateGraph(fileName, graphTitle, results, dataFinder.colourFinder)


class TitleFinder(HTMLParser):
    def __init__(self):
        HTMLParser.__init__(self)
        self.active = False
        self.title = None

    def handle_starttag(self, rawname, attrs):
        self.active = rawname.lower() == "title"

    def handle_data(self, content):
        if self.active:
            self.title = content
            self.active = False


class SummaryDataFinder:
    def __init__(self, location, apps, summaryFileName, basePath, defaultUsePie):
        self.diag = logging.getLogger("GenerateWebPages")
        self.location = location
        self.basePath = basePath
        self.summaryPageName = os.path.join(location, summaryFileName)
        self.appVersionInfo = {}
        self.appUsePie = {}
        self.appDirs = OrderedDict()
        self.appRuns = {}
        self.colourFinder, self.inputOptions = None, None
        if len(apps) > 0:
            self.colourFinder = testoverview.ColourFinder(apps[0][0].getCompositeConfigValue)
            self.inputOptions = apps[0][0].inputOptions
        appnames = set()
        for app, usePie, _ in apps:
            appnames.add(app.name)
            self.appUsePie[app.fullName()] = usePie
            repo = app.getBatchConfigValue("batch_result_repository")
            if repo:
                runDir = os.path.join(repo, "run_names")
                if os.path.isdir(runDir):
                    self.appRuns[app.fullName()] = runDir
            appDir = os.path.join(location, app.name)
            self.diag.info("Searching under " + repr(appDir))
            if os.path.isdir(appDir):
                self.appDirs[app.fullName()] = appDir

        if os.path.isdir(location):
            for dirName in os.listdir(location):
                if dirName not in appnames and not dirName.endswith("_history") and dirName not in ["images", "javascript", "jenkins_changes"]:
                    fullDir = os.path.join(location, dirName)
                    if os.path.isdir(fullDir):
                        appFullName = self.findFullName(fullDir)
                        if appFullName and appFullName not in self.appDirs:
                            self.appUsePie[appFullName] = defaultUsePie
                            self.appDirs[appFullName] = fullDir

    @staticmethod
    def findFullName(dirName):
        # All files have the application in the title attribute
        htmlFiles = glob(os.path.join(dirName, "test_*.html"))
        if len(htmlFiles) == 0:
            return
        anyFile = os.path.join(dirName, htmlFiles[0])
        finder = TitleFinder()
        finder.feed(open(anyFile).read())
        title = finder.title
        if title is not None:
            prefix = "est results for"
            pos = title.find(prefix) + len(prefix) + 1
            if pos != -1:
                endPos = title.find(" - ", pos)
                if endPos != -1:
                    return title[pos:endPos]

    def getTemplateFile(self):
        return self.ensureLocationFileExists("summary_template.html", "etc")

    def ensureLocationFileExists(self, fileName, dataDirName=""):
        locationFile = os.path.join(self.location, fileName)
        if not os.path.isfile(locationFile):
            plugins.ensureDirExistsForFile(locationFile)
            plugins.log.info("No file at '" + locationFile + "', copying default file from installation")
            includeSite, includePersonal = self.inputOptions.configPathOptions()
            srcFile = plugins.findDataPaths([fileName], includeSite, includePersonal, dataDirName)[-1]
            shutil.copyfile(srcFile, locationFile)
        return locationFile

    def getLink(self, path):
        relpath = plugins.relpath(self.summaryPageName, self.location, normalise=False)
        if "/" not in relpath:
            return path

        currLocation = self.location
        newParts = []
        for part in reversed(os.path.dirname(relpath).split("/")):
            if part == "..":
                currLocation, localName = os.path.split(currLocation)
                newParts.append(localName)
            else:
                newParts.append("..")
        return "/".join(newParts) + "/" + path

    def getGraphFile(self, appName, version):
        return os.path.join(self.location, self.getShortAppName(appName), "images", "GenerateGraphs_" + version + ".png")

    def getAppsWithVersions(self):
        appsWithVersions = OrderedDict()
        for appName, appDir in self.appDirs.items():
            self.diag.info("adding app " + appName + " with dir " + appDir)
            versionInfo = self.getVersionInfoFor(appDir)
            if versionInfo:
                self.appVersionInfo[appName] = versionInfo
                appsWithVersions[appName] = list(versionInfo.keys())
        return appsWithVersions

    def getShortAppName(self, fullName):
        appDir = self.appDirs[fullName]
        return os.path.basename(appDir)

    def getVersionInfoFor(self, appDir):
        versionDates = {}
        for path in sorted(glob(os.path.join(appDir, "test_*.html"))):
            fileName = os.path.basename(path)
            version, date, tag = parseFileName(fileName, self.diag)
            if version:
                overviewPage = os.path.join(appDir, self.getOverviewPageName(version))
                if os.path.isfile(overviewPage):
                    self.diag.info("Found file with version " + version)
                    versionDates.setdefault(version, {})
                    versionDates[version][(date, tag)] = path
        return versionDates

    def getOverviewPageName(self, version):
        return "test_" + version + ".html"

    def getOverviewPage(self, appName, version):
        return os.path.join(self.basePath, self.getShortAppName(appName), self.getOverviewPageName(version))

    def getAppRunDirectory(self, appName):
        if appName in self.appRuns:
            return self.appRuns.get(appName)
        else:
            # application in question might not be loaded right now
            allRunDirs = set(self.appRuns.values())
            if len(allRunDirs) == 1:
                return allRunDirs.pop()
            else:
                self.diag.info("No log dir found for " + appName + " stored are " + repr(self.appRuns))

    def getMostRecentDateAndTags(self):
        allInfo = {}
        for appName, appInfo in self.appVersionInfo.items():
            for version, versionData in appInfo.items():
                lastInfoPerEnv = self.getLastInfoPerEnvironment(list(versionData.keys()), self.getAppRunDirectory(appName))
                for envData, lastInfo in lastInfoPerEnv:
                    allInfo.setdefault(envData, []).append(lastInfo[0])
                self.diag.info("Most recent date for " + appName + " version " + version + " = " + repr(lastInfoPerEnv))
        mostRecent = set()
        for envData, lastInfoList in list(allInfo.items()):
            mostRecent.add(max(lastInfoList, key=self.getDateTagKey))
        return sorted(mostRecent, key=self.getDateTagKey)

    def getDateTagKey(self, info):
        return info[0], plugins.padNumbersWithZeroes(info[1])

    def usePieChart(self, appName):
        if self.appUsePie.get(appName):
            try:
                from .resultgraphs import PieGraph  # @UnusedImport
                return True
            except Exception as e:
                sys.stderr.write("Not producing pie charts for index pages: " + str(e) + "\n")
                self.appUsePie = {}
                return False  # if matplotlib isn't installed or is too old
        else:
            return False

    def extractLast(self, tags, count=1):
        if count == 1:
            return [max(tags, key=self.getDateTagKey)]
        else:
            return sorted(tags, key=self.getDateTagKey)[-count:]

    def getLastInfoPerEnvironment(self, allTags, runDir, count=1):
        if runDir is None:
            value = self.extractLast(allTags, count)
            return [((None, None), value)] if value is not None else []

        def getEnvironmentData(tag):
            date, actualTag = tag
            fullTag = time.strftime("%d%b%Y", date) + "_" + actualTag
            runEnv = getEnvironmentFromRunFiles([runDir], fullTag)
            return runEnv.get("JENKINS_URL"), runEnv.get("JOB_NAME")

        groupedData = {}
        for tag in allTags:
            envData = getEnvironmentData(tag)
            groupedData.setdefault(envData, []).append(tag)

        mostRecentDate = self.toDate(max(allTags, key=self.getDateTagKey)[0])
        oneDay = datetime.timedelta(days=1)
        allLastInfo = {}
        for envData, tags in groupedData.items():
            last = self.extractLast(tags, count)
            lastDate = self.toDate(last[-1][0])
            if last and mostRecentDate - lastDate <= oneDay:
                allLastInfo[envData] = last

        if len(allLastInfo) > 1 and (None, None) in allLastInfo:
            del allLastInfo[(None, None)]

        return list(allLastInfo.items())

    def toDate(self, timeStruct):
        return datetime.date.fromtimestamp(time.mktime(timeStruct))

    def getMax(self, values, ignoreItems, **kw):
        if ignoreItems == 0:
            return max(values, **kw)
        elif len(values) > ignoreItems:
            values.sort(**kw)
            return values[-1 - ignoreItems]

    def getLatestSummaries(self, appName, version):
        versionData = self.appVersionInfo[appName][version]
        summary, prevSummary = OrderedDict(), OrderedDict()
        infoPerEnv = self.getLastInfoPerEnvironment(list(versionData.keys()), self.getAppRunDirectory(appName), count=2)
        lastInfo, nextLastInfo = None, None
        for envData, lastInfoList in infoPerEnv:
            lastInfo = lastInfoList[-1]
            path = versionData[lastInfo]
            self.diag.info("Extracting summary information from " + path)
            self.extractSummary(path, summary)
            self.diag.info("For app " + appName + " version " + version + " environment " +
                           repr(envData) + ", found summary info " + repr(summary))
            if len(lastInfoList) == 2:
                nextLastInfo = lastInfoList[0]
                path = versionData[nextLastInfo]
                self.diag.info("Extracting previous summary information from " + path)
                self.extractSummary(path, prevSummary)
                self.diag.info("For app " + appName + " version " + version + " environment " +
                               repr(envData) + ", found previous summary info " + repr(prevSummary))
        self.diag.info("Last Info for version " + version + " = " +
                       repr(lastInfo) + ", previous = " + repr(nextLastInfo))
        return summary, lastInfo, prevSummary, nextLastInfo

    def getAllSummaries(self, appName, version):
        versionData = self.appVersionInfo[appName][version]
        allDates = list(versionData.keys())
        allDates.sort(key=self.getDateTagKey)
        summaries = [(time.strftime("%d%b%Y", currInfo[0]), self.extractSummary(
            versionData[currInfo], OrderedDict())) for currInfo in allDates]
        self.diag.info("All Summaries = " + repr(summaries))
        return summaries

    def extractSummary(self, datedFile, summary):
        for line in open(datedFile, errors="replace"):
            if line.strip().startswith("<H2>"):
                text = line.strip()[4:-5]  # drop the tags
                for cat, num in list(self.parseSummaryText(text).items()):
                    if cat in summary:
                        summary[cat] += num
                    else:
                        summary[cat] = num
                return summary
        return summary

    def parseSummaryText(self, text):
        words = text.split()[3:]  # Drop "Version: 12 tests"
        index = 0
        categories = []
        while index < len(words):
            try:
                count = int(words[index])
                categories.append(["", count])
            except ValueError:
                categories[-1][0] += words[index]
            index += 1
        self.diag.info("Category information is " + repr(categories))
        colourCount = OrderedDict()
        for colourKey in ["success", "knownbug", "performance", "failure", "incomplete"]:
            colourCount[colourKey] = 0
        for categoryName, count in categories:
            colourKey = self.getColourKey(categoryName)
            colourCount[colourKey] += count
        return colourCount

    def getColour(self, colourKey):
        return self.colourFinder.find(colourKey + "_bg")

    def getColourKey(self, categoryName):
        if categoryName == "succeeded":
            return "success"
        elif categoryName == "knownbugs":
            return "knownbug"
        else:
            for perfCat in ["faster", "slower", "memory+", "memory-", "larger", "smaller"]:
                if categoryName.startswith(perfCat):
                    return "performance"
            if categoryName in ["killed", "unrunnable", "cancelled", "abandoned"]:
                return "incomplete"
            else:
                return "failure"


class SummaryGenerator:
    def __init__(self):
        self.diag = logging.getLogger("GenerateWebPages")
        self.diag.info("Generating summary...")
        self.oldTimeFormat = "%Y%m%d_%H%M%S"
        self.timeFormat = self.oldTimeFormat + "_%f"

    def adjustLineForColours(self, line, dataFinder):
        mainPart = line.rsplit(";", 1)[0].rstrip()
        var, template = mainPart.rsplit(" ", 1)
        colourKey = template.split(".")[-1]
        return var + " " + dataFinder.getColour(colourKey) + ";\n"

    def getDateRangeText(self, info):
        dates = [i[0] for i in info]
        if len(dates) == 0:
            return ""
        firstDate = min(dates)
        lastDate = max(dates)
        text = " dated " + time.strftime("%d%b%Y", firstDate)
        if firstDate != lastDate:
            text += " - " + time.strftime("%d%b%Y", lastDate)
        return text

    def getRecentTagText(self, mostRecentInfo):
        if len(mostRecentInfo) > 2:
            return str(len(mostRecentInfo)) + " test runs" + self.getDateRangeText(mostRecentInfo)
        else:
            suffix = "s" if len(mostRecentInfo) > 1 else ""
            mostRecentTags = [i[1] for i in mostRecentInfo]
            return "test run" + suffix + " " + ", ".join(mostRecentTags)

    def generatePage(self, dataFinder, appsWithVersions, fileToUrl):
        jobLink = ""
        creationDate = testoverview.TitleWithDateStamp("").__str__().strip()
        if os.getenv("JENKINS_URL") and os.getenv("JOB_NAME") and os.getenv("BUILD_NUMBER"):
            jobPath = os.path.join(os.getenv("JENKINS_URL"), "job", os.getenv("JOB_NAME"), os.getenv("BUILD_NUMBER"))
            if jobPath:
                jobLink = "<br>(built by Jenkins job '" + os.getenv("JOB_NAME") + "', " + "<a href='" + \
                    jobPath + "'> " + "build number " + os.getenv("BUILD_NUMBER") + "</a>" + ")"

        summaryPageTimeStamp = dataFinder.summaryPageName + "." + plugins.startTimeString(self.timeFormat)
        with open(summaryPageTimeStamp, "w") as f:
            versionOrder = ["default"]
            appOrder = []
            mostRecentInfo = dataFinder.getMostRecentDateAndTags()
            mostRecentTags = [i[1] for i in mostRecentInfo]
            self.diag.info("Most recent results are from " + repr(mostRecentTags))
            cssColours = []
            for line in open(dataFinder.getTemplateFile()):
                if "<title>" in line:
                    f.write(line)
                elif "historical_report_colours" in line:
                    f.write(self.adjustLineForColours(line, dataFinder))
                else:
                    f.write(line)
                if "td.cell_" in line:
                    cssColours.append(line.rsplit("_", 1)[-1].split()[0])
                if "App order=" in line:
                    appOrder += self.extractOrder(line)
                if "Version order=" in line:
                    versionOrder += self.extractOrder(line)
                if "<h1" in line:
                    f.write("<h3 align=\"center\">(from " + self.getRecentTagText(mostRecentInfo) + ")</h3>\n")
                if "Insert table here" in line:
                    self.insertSummaryTable(f, dataFinder, mostRecentInfo, appsWithVersions,
                                            appOrder, versionOrder, cssColours)
                if "Insert footer here" in line:
                    f.write(creationDate + (jobLink if jobLink else ""))

        fileNames = self.getSortedFileNames(dataFinder.summaryPageName)
        self.linkOrCopy(fileNames[-1][1],
                        dataFinder.summaryPageName)
        self.cleanOldest(fileNames)

        plugins.log.info("wrote: '" + summaryPageTimeStamp + "'")
        if fileToUrl:
            url = convertToUrl(dataFinder.summaryPageName, fileToUrl)
            plugins.log.info("(URL is " + url + ")")

    def linkOrCopy(self, src, dst):
        if os.path.exists(dst):
            os.remove(dst)
        if os.name == "posix":
            os.symlink(os.path.basename(src), dst)
        else:
            shutil.copy(src, dst)

    def getTimeStampFromIndexFile(self, fileName):
        timeStampString = fileName.rsplit(".", 1)[-1]
        try:
            return datetime.datetime.strptime(timeStampString, self.timeFormat)
        except ValueError:
            try:
                return datetime.datetime.strptime(timeStampString, self.oldTimeFormat)
            except ValueError:
                pass  # Might not be the right format

    def cleanOldest(self, fileNames):
        numberOfFilesToKeep = 5
        anythingToRemove = len(fileNames) - numberOfFilesToKeep > 0
        if anythingToRemove:
            numberOfFilesToRemove = len(fileNames) - numberOfFilesToKeep

            # Don't remove newer files in order to mitigate the risk of
            # removing index.html links created by a job running at the
            # same time.
            deleteOlderThan = datetime.datetime.now() - datetime.timedelta(minutes=5)
            for _ in range(numberOfFilesToRemove):
                timeStamp, fileName = fileNames.pop(0)
                if timeStamp >= deleteOlderThan:
                    # fileNames is sorted by time stamps, so no
                    # more files can be removed.
                    break
                os.remove(fileName)

    def getSortedFileNames(self, root):
        allFileNames = glob(root + ".*")
        withTimeStamps = []
        for fileName in allFileNames:
            timeStamp = self.getTimeStampFromIndexFile(fileName)
            if timeStamp is not None:
                withTimeStamps.append((timeStamp, fileName))

        withTimeStamps.sort()
        return withTimeStamps

    def getOrderedVersions(self, predefined, info):
        fullList = sorted(info)
        versions = []
        for version in predefined:
            if version in fullList:
                versions.append(version)
                fullList.remove(version)
        return versions + fullList

    def padWithEmpty(self, versions, columnVersions, minColumnIndices):
        newVersions = []
        index = 0
        for version in versions:
            minIndex = minColumnIndices.get(version, 0)
            while index < minIndex:
                self.diag.info("Index = " + repr(index) + " but min index = " + repr(minIndex))
                newVersions.append("")
                index += 1
            while index in columnVersions and columnVersions[index] != version:
                newVersions.append("")
                index += 1
            newVersions.append(version)
            index += 1
        return newVersions

    def getMinColumnIndices(self, pageInfo, versionOrder):
        # We find the maximum column number a version has on any row,
        # which is equal to the minimum value it should be given in a particular row
        versionIndices = {}
        for rowInfo in list(pageInfo.values()):
            for index, version in enumerate(self.getOrderedVersions(versionOrder, rowInfo)):
                if version not in versionIndices or index > versionIndices[version]:
                    versionIndices[version] = index
        return versionIndices

    def getVersionsWithColumns(self, pageInfo):
        allVersions = reduce(operator.add, list(pageInfo.values()), [])
        return set([v for v in allVersions if allVersions.count(v) > 1])

    def createPieChart(self, dataFinder, resultSummary, summaryGraphName, version, lastInfo, oldResults):
        from .resultgraphs import PieGraph
        fracs = []
        colours = []
        tests = 0
        for colourKey, count in list(resultSummary.items()):
            if count:
                colour = dataFinder.getColour(colourKey)
                fracs.append(count)
                colours.append(colour)
                tests += count
        title = lastInfo[1] + " - " + str(tests) + " tests"
        pg = PieGraph(version, title, size=5)
        pg.pie(fracs, colours)
        summaryGraphFile = os.path.join(dataFinder.location, summaryGraphName)
        if oldResults:
            pg.save(summaryGraphFile, facecolor="#999999")
        else:
            pg.save(summaryGraphFile)

    def showResultAsOld(self, info, mostRecentInfo):
        # Only do this for timed results, i.e. nightjobs etc. From CI/Jenkins etc we can't really tell
        # - it depends on the schedule which we can't see and may not even exist
        if not self.isDate(info[1]):
            return False
        return all((info[0] != i[0] for i in mostRecentInfo))

    def isDate(self, tag):
        # Allow also for Jenkins format, 13Jan2006.1234
        return len(tag) >= 9 and tag[:2].isdigit() and tag[2:5].isalpha() and tag[5:9].isdigit()

    def getTrendImage(self, summary, prevSummary):
        text = self.getTrendText(summary, prevSummary)
        return "images/" + text + ".png"

    def getTrendText(self, summary, prevSummary):
        currSuccess = summary.get("success", 0)
        prevSuccess = prevSummary.get("success", 0)
        if currSuccess > prevSuccess:
            return "arrow_up"
        elif currSuccess < prevSuccess:
            return "arrow_down"
        elif summary.get("failure") > 0:
            return "red_arrow_across"
        else:
            return "green_arrow_across"

    def getTooltipForPrevious(self, prevSummary, prevTag):
        text = "Comparing to " + prevTag + " :\n"
        for colourKey, count in list(prevSummary.items()):
            if count:
                text += colourKey + "=" + str(count) + "\n"
        return text

    def getColourAttribute(self, colourKey, cssColours, dataFinder):
        if colourKey in cssColours:
            return 'class="cell_' + colourKey + '"'
        else:
            colour = dataFinder.getColour(colourKey)
            return 'bgcolor="' + colour + '"'

    def insertSummaryTable(self, file, dataFinder, mostRecentInfo, pageInfo, appOrder, versionOrder, cssColours):
        versionWithColumns = self.getVersionsWithColumns(pageInfo)
        self.diag.info("Following versions will be placed in columns " + repr(versionWithColumns))
        minColumnIndices = self.getMinColumnIndices(pageInfo, versionOrder)
        self.diag.info("Minimum column indices are " + repr(minColumnIndices))
        columnVersions = {}
        for appName in self.getOrderedVersions(appOrder, pageInfo):
            file.write('<tr class="application_row">\n')
            file.write('  <td class="application_name"><h3>' + appName + "</h3></td>\n")
            versions = pageInfo[appName]
            orderedVersions = self.getOrderedVersions(versionOrder, versions)
            self.diag.info("For " + appName + " found " + repr(orderedVersions))
            for columnIndex, version in enumerate(self.padWithEmpty(orderedVersions, columnVersions, minColumnIndices)):
                file.write('  <td class="version_table">')
                if version:
                    file.write('<table border="1" class="version_link"><tr>\n')
                    if version in versionWithColumns:
                        columnVersions[columnIndex] = version

                    resultSummary, lastInfo, prevResultSummary, nextLastInfo = dataFinder.getLatestSummaries(
                        appName, version)
                    oldResults = self.showResultAsOld(lastInfo, mostRecentInfo)
                    fileToLink = dataFinder.getOverviewPage(appName, version)
                    if dataFinder.usePieChart(appName):
                        summaryGraphName = "summary_pie_" + version + ".png"
                        self.createPieChart(dataFinder, resultSummary, summaryGraphName, version, lastInfo, oldResults)
                        file.write('    <td><a href="' + fileToLink + '"><img border=\"0\" src=\"' +
                                   summaryGraphName + '\"></a></td>\n')
                    else:
                        file.write('    <td><h3><a href="' + fileToLink + '">' + version + '</a></h3></td>\n')
                        for colourKey, count in list(resultSummary.items()):
                            if count:
                                file.write('    <td ' + self.getColourAttribute(colourKey,
                                                                                cssColours, dataFinder) + '><h3>')
                                if oldResults:
                                    # Highlight old data by putting it in a paler foreground colour
                                    file.write('<font color="#999999">' + str(count) + "</font>")
                                else:
                                    file.write(str(count))
                                file.write("</h3></td>\n")
                    if prevResultSummary:
                        image = self.getTrendImage(resultSummary, prevResultSummary)
                        dataFinder.ensureLocationFileExists(image)
                        tooltip = self.getTooltipForPrevious(prevResultSummary, nextLastInfo[1])
                        file.write('    <td class="arrow_cell"><img src="' +
                                   dataFinder.getLink(image) + '" title="' + tooltip + '"/></td>\n')
                    file.write("  </tr></table>")
                file.write("</td>\n")
            file.write("</tr>\n")

    def extractOrder(self, line):
        startPos = line.find("order=") + 6
        endPos = line.rfind("-->")
        return plugins.commasplit(line[startPos:endPos])




import os
from texttestlib import plugins

try:
    import storytext
    applicationEvent = storytext.applicationEvent
    applicationEventRename = storytext.applicationEventRename
except ImportError:
    def applicationEventRename(*args, **kw):
        pass

    def applicationEvent(*args, **kw):
        pass


# Compulsory responder to generate application events. Always present. See respond module
class ApplicationEventResponder(plugins.Responder):
    def __init__(self, *args):
        plugins.Responder.__init__(self)
        applicationEvent("test configuration to be read", "read")
        self.rerunTests = []

    def notifyLifecycleChange(self, test, state, changeDesc):
        if changeDesc.find("approved") != -1 or changeDesc.find("recalculated") != -1 or changeDesc.find("marked") != -1:
            # don't generate application events when a test is approved or recalculated or marked...
            return
        eventName = "test " + test.uniqueName + " to " + changeDesc
        if test in self.rerunTests:
            eventName += " rerun"
        category = test.uniqueName
        timeDelay = self.getTimeDelay()
        applicationEvent(eventName, category + " lifecycle", ["lifecycle"], timeDelay)

    def notifyAdd(self, test, initial):
        if initial and test.classId() == "test-case":
            eventName = "test " + test.uniqueName + " to be read"
            applicationEvent(eventName, test.uniqueName, [test.uniqueName + " lifecycle", "read", "lifecycle"])

    def notifyUniqueNameChange(self, test, newName):
        if test.classId() == "test-case":
            applicationEventRename("test " + test.uniqueName + " to", "test " + newName + " to",
                                   test.uniqueName, newName)

    def notifyRerunTriggered(self, test):
        self.rerunTests.append(test)

    def getTimeDelay(self):
        try:
            return int(os.getenv("TEXTTEST_FILEWAIT_SLEEP", 1))
        except ValueError:  # pragma: no cover - pathological case
            return 1

    def notifyBackgroundActionCompleted(self, *args):
        applicationEvent("completion of background actions", "background")

    def notifyAllRead(self, *args):
        applicationEvent("all tests to be read", "read", ["lifecycle"])

    def notifyAllComplete(self):
        applicationEvent("completion of test actions", "lifecycle")

    def notifyCloseDynamic(self, testArg, name):
        applicationEvent(name + " GUI to be closed")
